# -*- coding: UTF-8 -*-
from icv.utils import IS_PY3

try:
    if IS_PY3:
        from http.server import BaseHTTPRequestHandler, SimpleHTTPRequestHandler
    else:
        from SimpleHTTPServer import BaseHTTPRequestHandler, SimpleHTTPRequestHandler
except ImportError:
    raise ModuleNotFoundError

import io
from urllib.parse import urlparse
from ..exceptions import HttpParseException
import os
import re
from icv.utils import json_encode
from ..status_code import HTTP_STATUS_CODES
from http import HTTPStatus
import socket
from .context import Context


def api(ctx):
    data = ctx.request.data_json
    if data and "name" in data:
        print("yes!!!!!")
    ctx.response.send("post api")


def error(ctx):
    ctx.response.send({"error": "you are wrong"}, 500)


def ok_ke(ctx):
    ctx.response.send("post ok ke")


def hello(ctx):
    ctx.response.send("get hello")


def _init_routes():  # 初始化
    global routes
    global method_routes

    routes = {
        "/api": ("POST", api),
        "/error": (["POST", "PUT"], error),
        "/ok/ke": ("POST", ok_ke),
        "/hello": ("GET", hello),
    }
    method_routes = {
        "POST": [
            ("/api", api),
            ("/ok/ke", ok_ke)
        ],
        "GET": [
            ("/hello", hello)
        ]
    }


def add_route(action, methods, function):
    routes[action] = (methods, function)
    if isinstance(methods, str):
        method_routes[methods].append((action, function))
    for m in list(methods):
        if m not in method_routes:
            method_routes[m] = []
        method_routes[m].append((action, function))


class IcvHandler(SimpleHTTPRequestHandler):
    def _prepare_response(self, status_code=200, content_type="text/plain", **kwargs):
        self.send_response(status_code)
        self.send_header('Content-Type', content_type)
        for k in kwargs:
            self.send_header(k, kwargs[k])
        self.end_headers()

    def render(self, html, enc="utf-8"):
        print("#### render html:", html)
        if os.path.isfile(html):
            encoded = open(html, "r").read().encode(enc, 'surrogateescape')
        else:
            # encoded = html.encode(enc, 'surrogateescape')
            return self.error_404()
        f = io.BytesIO()
        f.write(encoded)
        f.seek(0)
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "text/html; charset=%s" % enc)
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        self.wfile.write(encoded)

    def error_404(self):
        self.error(404, "404 Not Found.")

    def error(self, status_code=500, errMsg=None):
        errMsg = errMsg if errMsg and isinstance(errMsg, str) else "Internal Server Error"
        self.response(errMsg, status_code=status_code)

    def response(self, data, status_code=200, **kwargs):
        if isinstance(data, dict):
            res = json_encode(data).encode("utf-8")
        elif isinstance(data, str):
            res = data.encode("utf-8")
        elif isinstance(data, bytes):
            res = data
        else:
            return self.error()

        content_type = None
        if isinstance(data, str):
            content_type = "text/plain;charset=utf-8"
        elif isinstance(data, dict):
            content_type = "application/json;charset=utf-8"
        else:
            if "Content-Type" in kwargs:
                content_type = kwargs["Content-Type"]
            elif "content-type" in kwargs:
                content_type = kwargs["content-type"]

        if content_type is not None:
            self._prepare_response(status_code=status_code, content_type=content_type, **kwargs)
        else:
            self._prepare_response(status_code=status_code, **kwargs)
        self.wfile.write(res)

    def _build_ctx(self):
        try:
            ctx = Context(self, self.command, dict(self.headers._headers))
            ctx.request.raw_path = self.path
            content_length = int(self.headers["Content-Length"]) if "Content-Length" in self.headers else 0
            post_data = self.rfile.read(content_length)
            ctx.request.data = post_data
            return ctx
        except HttpParseException as e:
            self.error(e.code, e.msg)
            self.finish()
            print("===> HttpParseException:", e.msg)
        except Exception as e:
            self.error()
            print("===> Exception:", e)

    def handle_one_request(self):
        try:
            self.raw_requestline = self.rfile.readline(65537)
            if len(self.raw_requestline) > 65536:
                self.requestline = ''
                self.request_version = ''
                self.command = ''
                self.send_error(HTTPStatus.REQUEST_URI_TOO_LONG)
                return
            if not self.raw_requestline:
                self.close_connection = True
                return
            if not self.parse_request():
                # An error code has been sent, just exit
                return
            mname = 'do_' + self.command
            if not hasattr(self, mname):
                self.send_error(
                    HTTPStatus.NOT_IMPLEMENTED,
                    "Unsupported method (%r)" % self.command)
                return

            up = urlparse(self.path)
            target = self._match(up.path.rstrip("/"))
            if target is None:
                if up.path.rstrip("/") == "":
                    self.response({
                        "up": "true",
                        "method": self.command,
                        "path": up.path,
                        "params": up.params,
                        "query": up.query,
                    })
                    return
                self.send_error(
                    HTTPStatus.NOT_FOUND,
                    "Not Found handler (%r)" % self.path
                )
                return

            ms, func, _ = target
            if isinstance(ms, str) and self.command != ms:
                self.send_error(
                    HTTPStatus.METHOD_NOT_ALLOWED,
                    "method not allowed (%r)" % self.command
                )
                return

            if isinstance(ms, (list, tuple)) and self.command not in ms:
                self.send_error(
                    HTTPStatus.METHOD_NOT_ALLOWED,
                    "method not allowed (%r)" % self.command
                )

            method = getattr(self, mname)
            method(func)
            self.wfile.flush()  # actually send the response if not already done.
        except socket.timeout as e:
            # a read or a write timed out.  Discard this connection
            self.log_error("Request timed out: %r", e)
            self.close_connection = True
            return

    def _match(self, rpath):
        for action in routes:
            handler_method, handler_func = routes[action]
            result = re.compile("^" + str(action) + "$").match(str(rpath))
            if result:
                return handler_method, handler_func, [x for x in result.groups()]
        # No match, return None
        return None

    def _filter(self, method=None):
        target = self._match(self.path)
        if target is None:
            self.error(404, HTTP_STATUS_CODES[404][0])
            self.finish()
            return None

        if method is None:
            return target

        handler_method, _, _ = target
        if isinstance(handler_method, str):
            if handler_method.upper() != str(method).upper():
                self.error(501, HTTP_STATUS_CODES[501][0])
                self.finish()
                return None
        else:
            if str(method).upper() not in handler_method:
                self.error(501, HTTP_STATUS_CODES[501][0])
                self.finish()
                return None

        return target

    def do_GET(self, handler=None):
        if handler is None:
            self.error(400, "handler not exist!")
            return

        handler(self._build_ctx())

    def do_PUT(self, handler=None):
        if handler is None:
            self.error(400, "handler not exist!")
            return

        handler(self._build_ctx())

    def do_DELETE(self, handler=None):
        if handler is None:
            self.error(400, "handler not exist!")
            return

        handler(self._build_ctx())

    def do_POST(self, handler=None):
        if handler is None:
            self.response("handler not exist!")
            return

        handler(self._build_ctx())



# -*- coding: UTF-8 -*-
from icv.utils import is_seq,is_str
from ..methods import HttpMethod
from .simple_http_server import IcvServer
from .simple_http_handler import IcvHandler
from . import simple_http_handler

class App:
    def __init__(self,host="0.0.0.0",port=9527,app_name="Icv Http Server",debug=False):
        self._host = host
        self._port = port
        self._app_name = app_name
        self._debug = debug
        self._server = IcvServer(host,port)
        simple_http_handler._init_routes()

    def route(self, action, methods=HttpMethod.methods):
        assert is_seq(methods) or is_str(methods)
        def decorate(function):
            simple_http_handler.add_route(action, methods, function)

        return decorate

    def start(self):
        print(">>>>>>> server now running at {}:{} <<<<<<<".format(self._host,self._port))
        self._server.run(IcvHandler)



# -*- coding: UTF-8 -*-
from ..voc import Voc
from ..coco import Coco
from ..core.sample import Sample
from icv.utils import reset_dir,make_empty_coco_anno
import json
import os

class VocConverter(object):
    def __init__(self,voc):
        assert isinstance(voc,Voc)
        self.voc = voc

    def to_coco(self,coco_root,split=None,reset=False):
        split = split if split is not None else self.voc.split
        if reset:
            reset_dir(coco_root)
        dist_anno_path, dist_image_path = Coco.reset_dir(coco_root,split)
        dist_anno_file = os.path.join(dist_anno_path,"%s.json" % split)

        anno_data = make_empty_coco_anno()
        json.dump(anno_data,open(dist_anno_file,"w+"))

        coco = Coco(
                    dist_image_path,
                    dist_anno_file,
                    keep_no_anno_image=self.voc.keep_no_anno_image,
                    one_index=self.voc.one_index,
                )
        coco.set_categories(self.voc.categories)

        image_id = 1

        for sample in self.voc.get_samples():
            assert isinstance(sample,Sample)
            coco.sample_db[image_id] = sample
            coco.ids.append(image_id)
            image_id += 1

        coco.save(coco_root,reset_dir=True,split=split)
        coco.init()
        return coco





# -*- coding: utf-8 -* -
import logging

def _validate_label_map(label_map):
    """Checks if a label map is valid.
    Args:
      label_map: StringIntLabelMap to validate.
    Raises:
      ValueError: if label map is invalid.
    """
    for item in label_map.item:
        if item.id < 0:
            raise ValueError('Label map ids should be >= 0.')
        if (item.id == 0 and item.name != 'background' and
                item.display_name != 'background'):
            raise ValueError('Label map id 0 is reserved for the background label')


def load_labelmap(path):
    """Loads label map proto.
    Args:
      path: path to StringIntLabelMap proto text file.
    Returns:
      a StringIntLabelMapProto
    """
    import tensorflow as tf
    from google.protobuf import text_format
    from ..core.protos import string_int_label_map_pb2

    with tf.gfile.GFile(path, 'r') as fid:
        label_map_string = fid.read()
        label_map = string_int_label_map_pb2.StringIntLabelMap()
        try:
            text_format.Merge(label_map_string, label_map)
        except text_format.ParseError:
            label_map.ParseFromString(label_map_string)
    _validate_label_map(label_map)
    return label_map


def convert_label_map_to_categories(label_map,
                                    max_num_classes,
                                    use_display_name=True):
    """Given label map proto returns categories list compatible with eval.
    This function converts label map proto and returns a list of dicts, each of
    which  has the following keys:
      'id': (required) an integer id uniquely identifying this category.
      'name': (required) string representing category name
        e.g., 'cat', 'dog', 'pizza'.
    We only allow class into the list if its id-label_id_offset is
    between 0 (inclusive) and max_num_classes (exclusive).
    If there are several items mapping to the same id in the label map,
    we will only keep the first one in the categories list.
    Args:
      label_map: a StringIntLabelMapProto or None.  If None, a default categories
        list is created with max_num_classes categories.
      max_num_classes: maximum number of (consecutive) label indices to include.
      use_display_name: (boolean) choose whether to load 'display_name' field as
        category name.  If False or if the display_name field does not exist, uses
        'name' field as category names instead.
    Returns:
      categories: a list of dictionaries representing all possible categories.
    """
    categories = []
    list_of_ids_already_added = []
    if not label_map:
        label_id_offset = 1
        for class_id in range(max_num_classes):
            categories.append({
                'id': class_id + label_id_offset,
                'name': 'category_{}'.format(class_id + label_id_offset)
            })
        return categories
    for item in label_map.item:
        if not 0 < item.id <= max_num_classes:
            logging.info(
                'Ignore item %d since it falls outside of requested '
                'label range.', item.id)
            continue
        if use_display_name and item.HasField('display_name'):
            name = item.display_name
        else:
            name = item.name
        if item.id not in list_of_ids_already_added:
            list_of_ids_already_added.append(item.id)
            categories.append({'id': item.id, 'name': name})
    return categories


def labelmap_to_categories(label_map_path, use_display_name=True):
    """Reads a label map and returns categories list compatible with eval.
    This function converts label map proto and returns a list of dicts, each of
    which  has the following keys:
      'id': an integer id uniquely identifying this category.
      'name': string representing category name e.g., 'cat', 'dog'.
    Args:
      label_map_path: Path to `StringIntLabelMap` proto text file.
      use_display_name: (boolean) choose whether to load 'display_name' field
        as category name.  If False or if the display_name field does not exist,
        uses 'name' field as category names instead.
    Returns:
      categories: a list of dictionaries representing all possible categories.
    """
    label_map = load_labelmap(label_map_path)
    max_num_classes = max(item.id for item in label_map.item)
    return convert_label_map_to_categories(label_map, max_num_classes,
                                           use_display_name)


def create_category_index(categories):
    """Creates dictionary of COCO compatible categories keyed by category id.
    Args:
      categories: a list of dicts, each of which has the following keys:
        'id': (required) an integer id uniquely identifying this category.
        'name': (required) string representing category name
          e.g., 'cat', 'dog', 'pizza'.
    Returns:
      category_index: a dict containing the same entries as categories, but keyed
        by the 'id' field of each category.
    """
    category_index = {}
    for cat in categories:
        category_index[cat['id']] = cat
    return category_index


def labelmap_to_category_index(label_map_path, use_display_name=True):
    """Reads a label map and returns a category index.
    Args:
      label_map_path: Path to `StringIntLabelMap` proto text file.
      use_display_name: (boolean) choose whether to load 'display_name' field
        as category name.  If False or if the display_name field does not exist,
        uses 'name' field as category names instead.
    Returns:
      A category index, which is a dictionary that maps integer ids to dicts
      containing categories, e.g.
      {1: {'id': 1, 'name': 'dog'}, 2: {'id': 2, 'name': 'cat'}, ...}
    """
    categories = labelmap_to_categories(label_map_path, use_display_name)
    return create_category_index(categories)



# -*- coding: utf-8 -* -
"""
模型输入前处理，以及模型预测后处理
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from abc import *


class Processor(abc.ABCMeta):

    @abstractmethod
    def pre_process(self, inputs):
        pass

    @abstractmethod
    def post_process(self, outputs):
        pass



# -*- coding: UTF-8 -*-
from collections import OrderedDict
from . import is_file, mkdir
import os
import time

def load_state_dict(module, state_dict, strict=False, logger=None):
    """Load state_dict to a module.

    This method is modified from :meth:`torch.nn.Module.load_state_dict`.
    Default value for ``strict`` is set to ``False`` and the message for
    param mismatch will be shown even if strict is False.

    Args:
        module (Module): Module that receives the state_dict.
        state_dict (OrderedDict): Weights.
        strict (bool): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``False``.
        logger (:obj:`logging.Logger`, optional): Logger to log the error
            message. If not specified, print function will be used.
    """
    import torch
    unexpected_keys = []
    own_state = module.state_dict()
    for name, param in state_dict.items():
        if name not in own_state:
            unexpected_keys.append(name)
            continue
        if isinstance(param, torch.nn.Parameter):
            # backwards compatibility for serialized parameters
            param = param.data

        try:
            own_state[name].copy_(param)
        except Exception:
            raise RuntimeError('While copying the parameter named {}, '
                               'whose dimensions in the model are {} and '
                               'whose dimensions in the checkpoint are {}.'
                               .format(name, own_state[name].size(),
                                       param.size()))
    missing_keys = set(own_state.keys()) - set(state_dict.keys())

    err_msg = []
    if unexpected_keys:
        err_msg.append('unexpected key in source state_dict: {}\n'.format(
            ', '.join(unexpected_keys)))
    if missing_keys:
        err_msg.append('missing keys in source state_dict: {}\n'.format(
            ', '.join(missing_keys)))
    err_msg = '\n'.join(err_msg)
    if err_msg:
        if strict:
            raise RuntimeError(err_msg)
        elif logger is not None:
            logger.warn(err_msg)
        else:
            print(err_msg)


def ckpt_load(model,
              filename,
              map_location=None,
              strict=False,
              logger=None):
    """Load checkpoint from a file or URI.

    Args:
        model (Module): Module to load checkpoint.
        filename (str): Either a filepath or URL or modelzoll://xxxxxxx.
        map_location (str): Same as :func:`torch.load`.
        strict (bool): Whether to allow different params for the model and
            checkpoint.
        logger (:mod:`logging.Logger` or None): The logger for error message.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    # load checkpoint from modelzoo or file or url
    import torch
    from torch.utils import model_zoo
    if filename.startswith('modelzoo://'):
        from torchvision.models.resnet import model_urls
        model_name = filename[11:]
        checkpoint = model_zoo.load_url(model_urls[model_name])
    elif filename.startswith(('http://', 'https://')):
        checkpoint = model_zoo.load_url(filename)
    else:
        if not is_file(filename):
            raise IOError('{} is not a checkpoint file'.format(filename))
        checkpoint = torch.load(filename, map_location=map_location)
    # get state_dict from checkpoint
    if isinstance(checkpoint, OrderedDict):
        state_dict = checkpoint
    elif isinstance(checkpoint, dict) and 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    else:
        raise RuntimeError(
            'No state_dict found in checkpoint file {}'.format(filename))
    # strip prefix of state_dict
    if list(state_dict.keys())[0].startswith('module.'):
        state_dict = {k[7:]: v for k, v in checkpoint['state_dict'].items()}
    # load state_dict
    if hasattr(model, 'module'):
        load_state_dict(model.module, state_dict, strict, logger)
    else:
        load_state_dict(model, state_dict, strict, logger)
    return checkpoint


def weights_to_cpu(state_dict):
    """Copy a model state_dict to cpu.

    Args:
        state_dict (OrderedDict): Model weights on GPU.

    Returns:
        OrderedDict: Model weights on GPU.
    """
    state_dict_cpu = OrderedDict()
    for key, val in state_dict.items():
        state_dict_cpu[key] = val.cpu()
    return state_dict_cpu


def ckpt_save(model, filename, optimizer=None, meta=None):
    """Save checkpoint to file.

    The checkpoint will have 3 fields: ``meta``, ``state_dict`` and
    ``optimizer``. By default ``meta`` will contain version and time info.

    Args:
        model (Module): Module whose params are to be saved.
        filename (str): Checkpoint filename.
        optimizer (:obj:`Optimizer`, optional): Optimizer to be saved.
        meta (dict, optional): Metadata to be saved in checkpoint.
    """
    import torch
    if meta is None:
        meta = {}
    elif not isinstance(meta, dict):
        raise TypeError('meta must be a dict or None, but got {}'.format(
            type(meta)))
    meta.update(time=time.asctime())

    mkdir(os.path.dirname(filename))
    if hasattr(model, 'module'):
        model = model.module

    checkpoint = {
        'meta': meta,
        'state_dict': weights_to_cpu(model.state_dict())
    }
    if optimizer is not None:
        checkpoint['optimizer'] = optimizer.state_dict()

    torch.save(checkpoint, filename)



# -*- coding: UTF-8 -*-

import re
import logging
from ..exceptions import HttpParseException

Log = logging.getLogger('simpleHttpServer.server')

class HttpRequest(object):
    def __init__(self, method, request_uri, protocol, headers):
        self.method = method
        self.request_uri = request_uri
        self.protocol = protocol
        self.headers = headers

    def __str__(self):
        return 'HttpRequest (method=%s, request_uri=%s, protocol=%s)' % \
               (self.method, self.request_uri, self.protocol)

    def is_range_requested(self):
        return 'Range' in self.headers

    def get_range(self):
        range_header_value = None

        if self.is_range_requested():
            range_header_value = self.headers['Range']

        if range_header_value:
            range_start, range_end = None, None

            range = re.findall(r'\d+', range_header_value)
            range_start = int(range[0])

            if len(range) > 1:
                range_end = int(range[1])

            return range_start, range_end

        return None, None


def parse_http_request(data):
    # guard, check input parameters
    if not data:
        error = 'Input parameter data must be provided.'
        Log.error(error)
        raise HttpParseException(error)

    data_lines = data.splitlines(False)

    request_line = data_lines[0]
    request_cmpts = request_line.split(' ')

    if len(request_cmpts) != 3:
        error = 'Cannot parse HTTP request line: %s' % request_line
        Log.error(error)
        raise HttpParseException(error)

    method, request_uri, protocol = request_cmpts[0], request_cmpts[1], request_cmpts[2]
    headers = {}

    for line in data_lines[1:]:
        if not line:
            break

        line_cmpts = line.split(': ')

        if len(line_cmpts) != 2:
            raise HttpParseException('Cannot parse HTTP header line: %s' % line)

        key, value = line_cmpts[0], line_cmpts[1]
        headers[key] = value

    return HttpRequest(method, request_uri, protocol, headers)




# -*- coding: UTF-8 -*-
import os
import shutil
from .dataset import IcvDataSet
from .core.meta import SampleMeta, AnnoMeta
from .core.sample import Sample, Anno
from .core.bbox import BBox
from .core.polys import Polygon
from ..utils import fcopy, is_file, is_empty, is_dir, make_empty_coco_anno, json_encode, encode_to_file, \
    decode_from_file
import numpy as np
from tqdm import tqdm


class Coco(IcvDataSet):
    def __init__(self, image_dir, anno_file, split=None, keep_no_anno_image=True, one_index=True, transform=None,
                 target_transform=None):
        assert os.path.isdir(image_dir), "param image_dir is not a dir!"
        assert os.path.exists(anno_file), "param anno_file is not exist!"

        self.image_dir = image_dir
        self.anno_file = anno_file
        self.split = split if split is not None else os.path.basename(anno_file).rsplit(".", 1)[0]

        self.keep_no_anno_image = keep_no_anno_image
        self.one_index = one_index
        self.transform = transform
        self.target_transform = target_transform

        self.init()
        super(Coco, self).__init__(self.ids, self.categories, self.keep_no_anno_image, one_index, False)

    def init(self):
        from pycocotools.coco import COCO
        self.coco = COCO(self.anno_file)
        self.categories = self.get_categories()
        self.ids = self.coco.getImgIds()

        self.sample_db = {}
        self.set_colormap()

        print("there have %d samples in COCO dataset" % len(self.ids))
        print("there have %d categories in COCO dataset" % len(self.categories))

    @property
    def is_seg_mode(self):
        if self.length == 0:
            return False

        for i in range(self.length):
            for anno in self.get_sample(self.ids[i]).annos:
                if anno.seg_mode_polys or anno.seg_mode_mask:
                    return True

        return False

    def concat(self, coco, out_dir=None, reset=False, new_split=None):
        assert isinstance(coco, Coco)
        split = self.split if new_split is None else new_split
        dist_anno_path, dist_image_path = Coco.reset_dir(out_dir, split, reset=reset)
        dist_anno_file = os.path.join(dist_anno_path, "%s.json" % split)

        # TODO: concat logic
        anno_dict = decode_from_file(self.anno_file)
        # copy image file
        fcopy([os.path.join(self.image_dir, self.coco.loadImgs(id)[0]['file_name']) for id in self.ids],
              dist_image_path)

        max_img_id = np.max(np.uint(self.ids))
        max_anno_id = np.max(list(self.coco.anns.keys()))
        max_cat_id = np.max(list(self.cat2id.values()))

        new_anno_id = max_anno_id + 1
        new_cat_id = max_cat_id + 1

        cat2id = dict(self.cat2id)
        image_id_map = {}
        for i, id in enumerate(coco.ids):
            sample = coco.get_sample(id)
            file_name = sample.meta.file_name
            dist_image_file = os.path.join(dist_image_path, file_name)
            new_filename = fcopy(os.path.join(coco.image_dir, file_name), dist_image_file)

            image_info = sample.meta.dict()
            image_info["id"] = max_img_id + i + 1
            image_info["file_name"] = new_filename

            anno_dict["images"].append(image_info)

            image_id_map[id] = image_info["image_id"]

            for anno in sample.annos:
                anno_meta = anno.meta.dict()
                anno_meta["id"] = new_anno_id
                new_anno_id += 1
                anno_meta["image_id"] = image_info["id"]
                if anno.label in cat2id:
                    anno_meta["category_id"] = cat2id[anno.label]
                else:
                    cat2id[anno.label] = new_cat_id
                    new_cat_id += 1
                    anno_meta["category_id"] = cat2id[anno.label]

                anno_dict["annotations"].append(anno_meta)

        anno_dict["categories"] = [{"supercategory": cat, "id": cat2id[cat], "name": cat} for cat in cat2id]

        encode_to_file(dist_anno_file, anno_dict)

        return Coco(
            image_dir=dist_image_path,
            anno_file=dist_anno_file,
            keep_no_anno_image=self.keep_no_anno_image,
            one_index=self.one_index,
            transform=self.transform,
            target_transform=self.target_transform
        )

    @staticmethod
    def reset_dir(dist_dir, split, reset=False):
        if not reset:
            assert is_dir(dist_dir)

        if reset and os.path.exists(dist_dir):
            shutil.rmtree(dist_dir)

        dist_anno_path = os.path.join(dist_dir, "annotations")
        dist_image_path = os.path.join(dist_dir, "images", split)

        for _path in [dist_anno_path, dist_image_path]:
            if reset or not is_dir(_path):
                os.makedirs(_path)

        return dist_anno_path, dist_image_path

    def get_categories(self):
        categories = []
        catIds = sorted(self.coco.getCatIds())
        for catid in catIds:
            categories.append(self.coco.cats[catid]["name"])
        self.set_categories(categories)
        return categories

    def get_sample(self, id):
        if id in self.sample_db:
            return self.sample_db[id]

        path = self.coco.loadImgs(id)[0]['file_name']
        image_file = os.path.join(self.image_dir, path)

        annos = []
        for ann in self.coco.imgToAnns[id]:
            xmin, ymin, width, height = ann["bbox"]
            cat = self.id2cat[ann["category_id"]]
            annos.append(Anno(
                bbox=BBox(xmin=xmin, ymin=ymin, xmax=xmin + width, ymax=ymin + height, label=cat),
                polys=Polygon(ann["segmentation"][0], label=cat) if "segmentation" in ann else None,
                label=cat,
                color=self.color_map[cat],
                meta=AnnoMeta({k: ann[k] for k in ann if k not in ["bbox", "category_id", "segmentation"]})
            ))

        img_info = self.coco.imgs[id]
        sample = Sample(
            name=path.rsplit(".", 1)[0],
            image=image_file,
            annos=annos,
            meta=SampleMeta(img_info)
        )
        sample.id = id
        self.sample_db[id] = sample
        return sample

    def save(self, output_dir, reset_dir=False, split=None):
        dist_anno_path, dist_image_path = Coco.reset_dir(output_dir, split, reset=reset_dir)
        dist_anno_path = os.path.join(dist_anno_path, "%s.json" % split) if is_dir(dist_anno_path) else dist_anno_path
        if is_file(self.anno_file) and not is_empty(self.image_dir):
            fcopy(self.anno_file, dist_anno_path)
            fcopy(self.image_dir, dist_image_path)
        else:
            self._write_sample(output_dir, split=split)

    def _write_sample(self, dist_dir, split=None):
        split = self.split if split is None else split
        dist_anno_path, dist_image_path = Coco.reset_dir(dist_dir, split)
        annotation = make_empty_coco_anno()

        anno_id = 1
        cats = {}
        if os.path.isdir(dist_dir):
            for id in self.ids:
                sample = self.get_sample(id)
                fcopy(sample.path, dist_image_path)
                img_height, img_width = sample.image.shape[:2]
                annotation["images"].append(
                    {
                        "id": id,
                        "file_name": os.path.basename(sample.path),
                        "width": img_width,
                        "height": img_height
                    }
                )

                for anno in sample.annos:
                    anno_dict = {}
                    anno_dict["bbox"] = [anno.bbox.xmin, anno.bbox.ymin, anno.bbox.width, anno.bbox.height]
                    if anno.seg_mode_polys:
                        anno_dict["segmentation"] = [anno.polys.exterior.flatten().tolist()]
                        anno_dict["area"] = anno.polys.area
                        anno_dict["iscrowd"] = 0
                    elif anno.seg_mode_mask:
                        anno_dict["segmentation"] = [anno.mask.to_ploygons().exterior.flatten().tolist()]
                        anno_dict["area"] = anno.mask.to_ploygons().area
                        anno_dict["iscrowd"] = 0

                    anno_dict["category_id"] = self.cat2id[anno.label]
                    anno_dict["image_id"] = id
                    anno_dict["id"] = anno_id
                    anno_id += 1

                    annotation["annotations"].append(anno_dict)

                    if anno.label not in cats:
                        cats[anno.label] = {"supercategory": "", "id": anno_dict["category_id"], "name": anno.label}

        annotation["categories"] = cats.values()
        encode_to_file(annotation, os.path.join(dist_anno_path, "%s.json" % split))

    # def _write_sample(self, anno_samples, dist_path):
    #     annotation = {
    #         "images": [],
    #         "annotations": [],
    #         "categories": [],
    #     }
    #
    #     anno_id = 0
    #     cats = {}
    #     for id, sample in enumerate(anno_samples):
    #         if sample.has_field("file_name"):
    #             continue
    #         img_height, img_width = sample.image.shape[:2]
    #         annotation["images"].append(
    #             {
    #                 "id": id,
    #                 "file_name": sample.get_field("file_name"),
    #                 "width": img_width,
    #                 "height": img_height
    #             }
    #         )
    #
    #         for bbox in sample.bbox_list:
    #             if bbox.has_field("label"):
    #                 continue
    #             anno = {
    #                 "bbox": [bbox.xmin, bbox.ymin, bbox.width, bbox.height]
    #             }
    #
    #             if bbox.has_field("segmentation"):
    #                 anno["segmentation"] = bbox.get_field("segmentation")
    #
    #             if bbox.has_field("area"):
    #                 anno["area"] = bbox.get_field("area")
    #
    #             if bbox.has_field("iscrowd"):
    #                 anno["iscrowd"] = bbox.get_field("iscrowd")
    #
    #             anno["category_id"] = self.cat2id[bbox.lable]
    #             anno["image_id"] = id
    #             anno["id"] = anno_id
    #             anno_id += 1
    #
    #             if bbox.lable not in cats:
    #                 cats[bbox.lable] = {"supercategory": "", "id": anno["category_id"], "name": bbox.lable}
    #
    #     annotation["categories"] = cats.values()
    #     json.dump(annotation, open(dist_path, "w"))

    def showAnns(self, id, with_bbox=True, with_seg=True, is_show=True, save_path=None):
        """
        Display the specified annotations.
        :param anns (array of object): annotations to display
        :return: None
        """
        sample = self.get_sample(id)
        return sample.vis(with_bbox=with_bbox, with_seg=with_seg, is_show=is_show, save_path=save_path)

    def vis(self, id=None, with_bbox=True, with_seg=True, is_show=False, save_dir=None, reset_dir=False):
        if save_dir is not None:
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            elif reset_dir:
                shutil.rmtree(save_dir)
                os.makedirs(save_dir)

        if id is not None:
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            return self.showAnns(id, with_bbox=with_bbox, with_seg=with_seg, is_show=is_show, save_path=save_path)

        image_vis = []
        for id in tqdm(self.ids):
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            image = self.showAnns(id, with_bbox=with_bbox, with_seg=with_seg, is_show=False, save_path=save_path)
            image_vis.append(image)
        return image_vis



# -*- coding: UTF-8 -*-
from .service.client import DetectorClient

try:
    from .tf_obj import TfObjectDetector
except Exception as e:
    pass

try:
    from .mb import MbDetector
except Exception as e:
    pass

try:
    from .mmd import MmdetDetector
except Exception as e:
    pass

# __all__ = ['DetectorClient', 'TfObjectDetector', 'MbDetector', 'MmdetDetector']



# -*- coding: UTF-8 -*-
from ..vis.color import STANDARD_COLORS
from ..utils import is_seq, is_dir, mkdir
from ..image.transforms import imcrop, imresize
from .classify import Classify
from .core import BBox, Image
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from terminaltables import AsciiTable
import shutil
import random
import copy
import os
from time import time


class IcvDataSet(object):
    __metaclass__ = ABCMeta

    def __init__(self, ids=None, categories=None, keep_no_anno_image=True, one_index=False, build_cat=True):
        self.ids = ids if ids else []
        self.categories = categories if categories else []
        self.keep_no_anno_image = keep_no_anno_image
        self.one_index = one_index
        self.num_classes = len(self.categories)

        if build_cat:
            self._build()

        if not hasattr(self, "color_map") or self.color_map is None or len(self.color_map) == 0:
            self.set_colormap()

    def _build(self):
        if self.one_index:
            self.id2cat = {i + 1: cat for i, cat in enumerate(self.categories)}
            self.cat2id = {cat: i + 1 for i, cat in enumerate(self.categories)}
        else:
            self.id2cat = {i: cat for i, cat in enumerate(self.categories)}
            self.cat2id = {cat: i for i, cat in enumerate(self.categories)}

        self.id2index = {id: i for i, id in enumerate(self.ids)}
        self.index2id = {i: id for i, id in enumerate(self.ids)}

    def get_img_info(self, index):
        img_id = self.ids[index]
        return img_id

    def set_colormap(self, color_list_or_map=None):
        assert color_list_or_map is None or isinstance(color_list_or_map, dict) or is_seq(color_list_or_map)
        if color_list_or_map is not None and len(self.categories) > 0:
            if isinstance(color_list_or_map, dict):
                self.color_map = {c: color_list_or_map[c] for c in color_list_or_map if c in self.categories}
                self.color_map.update({
                    c: random.choice(STANDARD_COLORS)
                    for c in self.categories if c not in self.color_map
                })
            elif is_seq(color_list_or_map) and len(color_list_or_map) > 0:
                self.color_map = {self.categories[i]: color_list_or_map[i % len(color_list_or_map)] for i in
                                  range(len(self.categories))}
            return
        self.color_map = {}
        _colors = STANDARD_COLORS
        # random.shuffle(_colors)
        for i, cat in enumerate(self.categories):
            self.color_map[cat] = _colors[i % len(_colors)]

    @property
    def length(self):
        return len(self)

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, item):
        return self.get_sample(self.index2id[item])

    def get_categories(self):
        return self.categories

    def set_categories(self, categories):
        assert is_seq(categories)
        self.categories = list(categories)

        self.id2cat = {}
        self.cat2id = {}
        for i, cate in enumerate(self.categories):
            if self.one_index:
                self.id2cat[i + 1] = cate
                self.cat2id[cate] = i + 1
            else:
                self.id2cat[i] = cate
                self.cat2id[cate] = i

    def get_samples(self, ids=None):
        self.samples = []
        ids = ids if ids is not None else self.ids
        for id in ids:
            sample = self.get_sample(id)
            if sample.count == 0 and not self.keep_no_anno_image:
                continue
            self.samples.append(sample)

        return self.samples

    def get_sample(self, id):
        pass

    def get_groundtruth(self, id):
        return self.get_sample(id)

    @abstractmethod
    def vis(self, save_dir=None):
        pass

    @abstractmethod
    def sub(self, count=0, ratio=0, shuffle=True, output_dir=None):
        if count <= 0:
            count = round(self.length * ratio)

        count = min(self.length, count)
        if shuffle:
            ids = random.sample(self.ids, count)
        else:
            ids = self.ids[:count]

        dataset = copy.deepcopy(self)
        dataset.ids = ids
        dataset.categories = dataset.get_categories()

        if output_dir is not None:
            dataset.save(output_dir)

        return dataset

    @abstractmethod
    def save(self, output_dir, reset_dir=False, split=None):
        pass

    @abstractmethod
    def concat(self, dataset, out_dir, reset=False, new_split=None):
        pass

    @abstractmethod
    def crop_bbox_for_classify(self, output_dir, reset=False, split="train", pad=0, resize=None):
        assert resize is None or (is_seq(resize) and len(resize) == 2)
        if reset and is_dir(output_dir):
            shutil.rmtree(output_dir)
        if not is_dir(output_dir):
            mkdir(output_dir)
        samples = self.get_samples()

        tmp_dir = os.path.join(output_dir, str(time()))
        mkdir(tmp_dir)

        classify_dataset = Classify(root_dir=output_dir, categories=self.categories)
        for sample in samples:
            _bboxes = []
            _cats = []
            for anno in sample.annos:
                if isinstance(anno.bbox, BBox):
                    _bbox = anno.bbox.copy()
                    _bbox.extend(pad, sample.shape)
                    _bboxes.append(_bbox.bbox)
                    _cats.append(anno.label)
            patches = imcrop(sample.image, _bboxes)
            assert len(patches) == len(_cats)

            for i, p in enumerate(patches):
                if resize is not None:
                    p = imresize(p, resize)
                imgobj = Image(img=p, ext=sample.ext, name="")
                classify_dataset.img2cat[imgobj] = _cats[i]
                classify_dataset.img2set[imgobj] = split

        classify_dataset.save(output_dir, rename=True)
        shutil.rmtree(tmp_dir)

    @abstractmethod
    def statistic(self, print_log=False):
        samples = self.get_samples()

        sta = dict(overview=dict(), detail=dict())
        sta["overview"] = {
            "image_num": self.length,
            "class_num": len(self.categories),
            "bbox_num": sum([sample.count for sample in samples]),
            "min_width": min([sample.width for sample in samples]),
            "max_width": max([sample.width for sample in samples]),
            "min_height": min([sample.height for sample in samples]),
            "max_height": max([sample.height for sample in samples]),
            "same_size": len(set([(sample.height, sample.width) for sample in samples])) == 1,
            "ratio_distribution": OrderedDict(),
        }

        sta["detail"] = {
            cat: {
                "num": 0,
                "ratio_distribution": OrderedDict(),
            }
            for cat in self.categories
        }

        for sample in samples:
            s = sample.statistics()
            for label in s["cats"]:
                sta["detail"][label]["num"] = s["cats"][label]

            for label in s["ratios"]:
                for r in s["ratios"][label]:
                    ratio = str(r)
                    if ratio not in sta["overview"]["ratio_distribution"]:
                        sta["overview"]["ratio_distribution"][ratio] = 0
                    sta["overview"]["ratio_distribution"][ratio] += s["ratios"][label][r]

                    if ratio not in sta["detail"][label]["ratio_distribution"]:
                        sta["detail"][label]["ratio_distribution"][ratio] = 0
                    sta["detail"][label]["ratio_distribution"][ratio] += s["ratios"][label][r]

        if print_log:
            overview_header = ["image_num", "class_num", "bbox_num", "same_image_size",
                               "min_image_width", "max_image_width", "min_image_height", "max_image_height"]
            overview_table_data = [
                overview_header,
                [
                    sta["overview"]["image_num"], sta["overview"]["class_num"], sta["overview"]["bbox_num"],
                    sta["overview"]["same_size"], sta["overview"]["min_width"], sta["overview"]["max_width"],
                    sta["overview"]["min_height"], sta["overview"]["max_height"],
                ]
            ]
            overview_table = AsciiTable(overview_table_data)
            overview_table.inner_footing_row_border = True

            ratios = sorted(list(sta["overview"]["ratio_distribution"].keys()))

            ratio_distri_header = ["bbox ratio (h/w)"] + ratios
            ratio_distri_table_data = [
                ratio_distri_header,
                ["-"] + [sta["overview"]["ratio_distribution"][r] for r in ratios]
            ]

            ratio_distri_table = AsciiTable(ratio_distri_table_data)
            ratio_distri_table.inner_footing_row_border = True

            cats_detail_header = ["bbox ratio (h/w) \ class"] + self.categories
            cats_detail_table_data = [cats_detail_header]
            cats_detail_table_data.append(["num"] + [sta["detail"][cat]["num"] for cat in self.categories])
            for r in ratios:
                cats_detail_table_data.append(
                    [r] + [sta["detail"][cat]["ratio_distribution"].get(r, 0) for cat in self.categories])

            cats_detail_table = AsciiTable(cats_detail_table_data)
            cats_detail_table.inner_footing_row_border = True

            print("overview:")
            print(overview_table.table)

            print("\nratio distribution:")
            print(ratio_distri_table.table)

            print("\nclass detail:")
            print(cats_detail_table.table)

        return sta



# -*- coding: UTF-8 -*-

HTTP_STATUS_CODES = {
    100: ('100 Continue', 'Request received, please continue'),
    101: ('101 Switching Protocols',
          'Switching to new protocol; obey Upgrade header'),

    200: ('200 OK', 'Request fulfilled, document follows'),
    201: ('201 Created', 'Document created, URL follows'),
    202: ('202 Accepted',
          'Request accepted, processing continues off-line'),
    203: ('203 Non-Authoritative Information', 'Request fulfilled from cache'),
    204: ('204 No Content', 'Request fulfilled, nothing follows'),
    205: ('205 Reset Content', 'Clear input form for further input.'),
    206: ('206 Partial Content', 'Partial content follows.'),

    300: ('300 Multiple Choices',
          'Object has several resources -- see URI list'),
    301: ('301 Moved Permanently', 'Object moved permanently -- see URI list'),
    302: ('302 Found', 'Object moved temporarily -- see URI list'),
    303: ('303 See Other', 'Object moved -- see Method and URL list'),
    304: ('304 Not Modified',
          'Document has not changed since given time'),
    305: ('305 Use Proxy',
          'You must use proxy specified in Location to access this '
          'resource.'),
    307: ('307 Temporary Redirect',
          'Object moved temporarily -- see URI list'),

    400: ('400 Bad Request',
          'Bad request syntax or unsupported method'),
    401: ('401 Unauthorized',
          'No permission -- see authorization schemes'),
    402: ('402 Payment Required',
          'No payment -- see charging schemes'),
    403: ('403 Forbidden',
          'Request forbidden -- authorization will not help'),
    404: ('404 Not Found', 'Nothing matches the given URI'),
    405: ('405 Method Not Allowed',
          'Specified method is invalid for this resource.'),
    406: ('406 Not Acceptable', 'URI not available in preferred format.'),
    407: ('407 Proxy Authentication Required', 'You must authenticate with '
                                           'this proxy before proceeding.'),
    408: ('408 Request Timeout', 'Request timed out; try again later.'),
    409: ('409 Conflict', 'Request conflict.'),
    410: ('410 Gone',
          'URI no longer exists and has been permanently removed.'),
    411: ('411 Length Required', 'Client must specify Content-Length.'),
    412: ('412 Precondition Failed', 'Precondition in headers is false.'),
    413: ('413 Request Entity Too Large', 'Entity is too large.'),
    414: ('414 Request-URI Too Long', 'URI is too long.'),
    415: ('415 Unsupported Media Type', 'Entity body in unsupported format.'),
    416: ('416 Requested Range Not Satisfiable',
          'Cannot satisfy request range.'),
    417: ('417 Expectation Failed',
          'Expect condition could not be satisfied.'),

    500: ('500 Internal Server Error', 'Server got itself in trouble'),
    501: ('501 Not Implemented',
          'Server does not support this operation'),
    502: ('502 Bad Gateway', 'Invalid responses from another server/proxy.'),
    503: ('503 Service Unavailable',
          'The server cannot process the request due to a high load'),
    504: ('504 Gateway Timeout',
          'The gateway server did not receive a timely response'),
    505: ('505 HTTP Version Not Supported', 'Cannot fulfill request.'),
}


# -*- coding: UTF-8 -*-
from xml.etree import ElementTree as etree
from lxml.etree import Element, SubElement, ElementTree
import os
import shutil


def load_voc_anno(anno_path, filter_empty=False):
    """Recursively parses XML contents to python dict.

    We assume that `object` tags are the only ones that can appear
    multiple times at the same level of a tree.

    Args:
      xml: xml tree obtained by parsing XML file contents using lxml.etree

    Returns:
      Python dictionary holding XML contents.
    """

    def _load_anno(xml):
        if not xml:
            return {xml.tag: xml.text}
        result = {}
        for child in xml:
            child_result = _load_anno(child)
            if child.tag != 'object':
                result[child.tag] = child_result[child.tag]
            else:
                if child.tag not in result:
                    result[child.tag] = []
                result[child.tag].append(child_result[child.tag])
        return {xml.tag: result}

    with open(anno_path, 'r') as fid:
        xml_str = fid.read()
    if "<bndbox>" not in xml_str and filter_empty:
        return {}
    xml = etree.fromstring(xml_str)
    return _load_anno(xml)


def save_voc_anno(anno_data, anno_dist_path):
    assert "folder" in anno_data
    assert "filename" in anno_data
    assert "size" in anno_data
    assert "width" in anno_data["size"]
    assert "height" in anno_data["size"]
    assert "depth" in anno_data["size"]
    assert "objects" in anno_data
    assert "segmented" in anno_data

    segmented = str(anno_data["segmented"])
    pose = "Unspecified"
    truncated = "0"
    difficult = "0"

    root = Element("annotation")
    SubElement(root, 'folder').text = anno_data["folder"]
    SubElement(root, 'filename').text = anno_data["filename"]

    source = SubElement(root, 'source')
    SubElement(source, 'database').text = "The VOC2007 Database"
    SubElement(source, 'annotation').text = "PASCAL VOC2007"
    SubElement(source, 'image').text = "flickr"

    size = SubElement(root, 'size')
    SubElement(size, 'width').text = str(anno_data["size"]["width"])
    SubElement(size, 'height').text = str(anno_data["size"]["height"])
    SubElement(size, 'depth').text = str(anno_data["size"]["depth"])

    SubElement(root, 'segmented').text = segmented

    for object in anno_data["objects"]:
        obj = SubElement(root, 'object')
        SubElement(obj, 'name').text = object["name"]
        SubElement(obj, 'pose').text = str(object["pose"]) if "pose" in object else pose
        SubElement(obj, 'truncated').text = str(object["truncated"]) if "truncated" in object else truncated
        SubElement(obj, 'difficult').text = str(object["difficult"]) if "difficult" in object else difficult
        bndbox = SubElement(obj, 'bndbox')
        SubElement(bndbox, 'xmin').text = str(object["bndbox"]["xmin"])
        SubElement(bndbox, 'ymin').text = str(object["bndbox"]["ymin"])
        SubElement(bndbox, 'xmax').text = str(object["bndbox"]["xmax"])
        SubElement(bndbox, 'ymax').text = str(object["bndbox"]["ymax"])

    tree = ElementTree(root)
    tree.write(anno_dist_path, encoding='utf-8', pretty_print=True)


def make_empty_coco_anno(**kwargs):
    empty_anno = {
        "images": [],
        "type": "instances",
        "annotations": [],
        "categories": [],
    }

    for k in kwargs:
        empty_anno[k] = kwargs[k]

    return empty_anno


def make_empty_voc_anno(**kwargs):
    empty_anno = {
        "folder": "",
        "filename": "",
        "source": {
            "database": "The VOC2012 Database",
            "annotation": "PASCAL VOC2007",
            "image": "flickr",
        },
        "size": {
            "width": "0",
            "height": "0",
            "depth": "0"
        },
        "segmented": "0",
        "objects": {},
    }

    if "folder" in kwargs:
        empty_anno["folder"] = kwargs["folder"]

    if "filename" in kwargs:
        empty_anno["filename"] = kwargs["filename"]

    if "width" in kwargs:
        empty_anno["size"]["width"] = kwargs["width"]

    if "height" in kwargs:
        empty_anno["size"]["height"] = kwargs["height"]

    if "depth" in kwargs:
        empty_anno["size"]["depth"] = kwargs["depth"]

    return empty_anno


def reset_voc_dir(dist_dir, with_segment=True):
    if not os.path.exists(dist_dir):
        os.makedirs(dist_dir)
    shutil.rmtree(dist_dir)

    dist_image_path = os.path.join(dist_dir, "JPEGImages")
    dist_anno_path = os.path.join(dist_dir, "Annotations")
    dist_imageset_path = os.path.join(dist_dir, "ImageSets", "Main")

    dist_segmentation_imageset_path = os.path.join(dist_dir, "Segmentation", "Main")
    dist_segmentation_class_image_path = os.path.join(dist_dir, "SegmentationClass")
    dist_segmentation_object_image_path = os.path.join(dist_dir, "SegmentationObject")

    os.makedirs(dist_image_path)
    os.makedirs(dist_anno_path)
    os.makedirs(dist_imageset_path)

    if with_segment:
        os.makedirs(dist_segmentation_imageset_path)
        os.makedirs(dist_segmentation_class_image_path)
        os.makedirs(dist_segmentation_object_image_path)

        return dist_image_path, dist_anno_path, dist_imageset_path, \
               dist_segmentation_imageset_path, dist_segmentation_class_image_path, dist_segmentation_object_image_path

    return dist_image_path, dist_anno_path, dist_imageset_path


def reset_coco_dir(dist_dir):
    if not os.path.exists(dist_dir):
        os.makedirs(dist_dir)
    shutil.rmtree(dist_dir)

    anno_dir = os.path.join(dist_dir, "annotations")
    train_image_dir = os.path.join(dist_dir, "train")
    test_image_dir = os.path.join(dist_dir, "test")
    val_image_dir = os.path.join(dist_dir, "val")

    os.makedirs(anno_dir)
    os.makedirs(train_image_dir)
    os.makedirs(test_image_dir)
    os.makedirs(val_image_dir)

    return anno_dir, train_image_dir, test_image_dir, val_image_dir



# -*- coding: UTF-8 -*-
import numpy as np
import cv2
from . import FLIP_TOP_BOTTOM,FLIP_LEFT_RIGHT

class MaskList(object):
    """
    Binary Mask List
    """
    def __init__(self,masks):
        if isinstance(masks,MaskList):
            self.masks = masks.masks
        elif isinstance(masks, (list, tuple)):
            import pycocotools.mask as mask_utils
            if isinstance(masks[0],np.ndarray):
                assert all([m.shape == masks[0].shape for m in masks])
                masks = np.array(masks)
                if len(masks.shape) == 2:
                    self.masks = masks[None]
            elif isinstance(masks[0], dict) and "counts" in masks[0]:
                masks = mask_utils.decode(masks)  # [h, w, n]
                self.masks = masks[...,[2,0,1]]

            assert len(self.masks.shape) == 3
            self.count = self.masks.shape[0]
            self.size = (self.masks.shape[1],self.masks.shape[2])

    def transpose(self, method):
        import torch
        dim = 1 if method == FLIP_TOP_BOTTOM else 2
        flipped_masks = torch.Tensor(self.masks).flip(dim).detach().numpy()
        return MaskList(flipped_masks)

    def crop(self, box):
        assert isinstance(box, (list, tuple)), str(type(box))
        # box is assumed to be xyxy
        current_width, current_height = self.size
        xmin, ymin, xmax, ymax = [round(float(b)) for b in box]

        assert xmin <= xmax and ymin <= ymax, str(box)
        xmin = min(max(xmin, 0), current_width - 1)
        ymin = min(max(ymin, 0), current_height - 1)

        xmax = min(max(xmax, 0), current_width)
        ymax = min(max(ymax, 0), current_height)

        xmax = max(xmax, xmin + 1)
        ymax = max(ymax, ymin + 1)

        width, height = xmax - xmin, ymax - ymin
        cropped_masks = self.masks[:, ymin:ymax, xmin:xmax]
        cropped_size = width, height
        return MaskList(cropped_masks)

    def resize(self, size):
        import torch
        try:
            iter(size)
        except TypeError:
            assert isinstance(size, (int, float))
            size = size, size
        width, height = map(int, size)

        assert width > 0
        assert height > 0

        # Height comes first here!
        # TODO: 双线性插值的包存在
        resized_masks = torch.nn.functional.interpolate(
            input=self.masks[None],
            size=(height, width),
            mode="bilinear",
            align_corners=False,
        )[0].type_as(self.masks)
        return MaskList(resized_masks)

    def _findContours(self):
        contours = []
        masks = self.masks.detach().numpy()
        for mask in masks:
            mask = cv2.UMat(mask)
            contour, hierarchy = cv2.findContours(
                mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_L1
            )

            reshaped_contour = []
            for entity in contour:
                assert len(entity.shape) == 3
                assert entity.shape[1] == 1, "Hierarchical contours are not allowed"
                reshaped_contour.append(entity.reshape(-1).tolist())
            contours.append(reshaped_contour)
        return contours

    def __len__(self):
        return len(self.masks)

    def __getitem__(self, index):
        # Probably it can cause some overhead
        # but preserves consistency
        masks = self.masks[index].clone()
        return MaskList(masks)

    def __iter__(self):
        return iter(self.masks)

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "num_instances={}, ".format(len(self.masks))
        s += "image_width={}, ".format(self.size[1])
        s += "image_height={})".format(self.size[0])
        return s





# -*- coding: UTF-8 -*-
import numpy as np

class Keypoint(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    @property
    def x_int(self):
        """
        Return the keypoint's x-coordinate, rounded to the closest integer.
        Returns
        -------
        result : int
            Keypoint's x-coordinate, rounded to the closest integer.
        """
        return int(np.round(self.x))

    @property
    def y_int(self):
        """
        Return the keypoint's y-coordinate, rounded to the closest integer.
        Returns
        -------
        result : int
            Keypoint's y-coordinate, rounded to the closest integer.
        """
        return int(np.round(self.y))

    def copy(self, x=None, y=None):
        """
        Create a shallow copy of the Keypoint object.
        Parameters
        ----------
        x : None or number, optional
            Coordinate of the keypoint on the x axis.
            If ``None``, the instance's value will be copied.
        y : None or number, optional
            Coordinate of the keypoint on the y axis.
            If ``None``, the instance's value will be copied.
        Returns
        -------
        imgaug.Keypoint
            Shallow copy.
        """
        return self.deepcopy(x=x, y=y)

    def deepcopy(self, x=None, y=None):
        """
        Create a deep copy of the Keypoint object.
        Parameters
        ----------
        x : None or number, optional
            Coordinate of the keypoint on the x axis.
            If ``None``, the instance's value will be copied.
        y : None or number, optional
            Coordinate of the keypoint on the y axis.
            If ``None``, the instance's value will be copied.
        Returns
        -------
        imgaug.Keypoint
            Deep copy.
        """
        x = self.x if x is None else x
        y = self.y if y is None else y
        return Keypoint(x=x, y=y)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Keypoint(x=%.8f, y=%.8f)" % (self.x, self.y)



# -*- coding: UTF-8 -*-
from icv.utils import is_valid_url, json_encode
from ..methods import HttpMethod
from .request import IcvHttpRequest


class IcvHttpClient(object):
    def __init__(self, url, data=None, params=None, headers=None, timeout=3000):
        if not is_valid_url(url):
            raise Exception("request url is invalid")

        self.method = "GET"
        self.url = url
        self.data = data
        self.params = params if params else {}
        self.headers = headers if headers else {}
        self.timeout = timeout

    def set_method(self, method):
        method = str(method).upper()
        if method not in HttpMethod:
            raise Exception("request method invalid")
        self.method = method

    def set_headers(self, headers):
        self.headers = headers if headers else {}

    def header(self, key, value):
        self.headers[key] = value

    def set_data(self, data):
        self.data = data

    def set_params(self, params):
        self.params = params if params else {}

    def set_timeout(self, timeout):
        self.timeout = timeout

    def do_request(self):
        req = IcvHttpRequest(self.method, self.url, data=self.data, params=self.params, headers=self.headers,
                             timeout=self.timeout)
        return req.send()

    def post(self, json=False, form=False):
        if json:
            if isinstance(self.data, dict):
                self.data = json_encode(self.data).encode("utf-8")
            self.header("Content-Type", "application/json")
        elif form:
            if isinstance(self.data, dict):
                self.data = "&".join(["{}={}".format(k, self.data[k]) for k in self.data])
            self.header("Content-Type", "application/x-www-form-urlencoded")

        self.method = HttpMethod.POST
        return self.do_request()

    def get(self):
        self.method = HttpMethod.GET

        return self.do_request()

    def put(self, json=False, form=False):
        if json:
            if isinstance(self.data, dict):
                self.data = json_encode(self.data).encode("utf-8")
            self.header("Content-Type", "application/json")
        elif form:
            if isinstance(self.data, dict):
                self.data = "&".join(["{}={}".format(k, self.data[k]) for k in self.data])
            self.header("Content-Type", "application/x-www-form-urlencoded")

        self.method = HttpMethod.PUT
        return self.do_request()

    def delete(self):
        self.method = HttpMethod.DELETE
        return self.do_request()

    def options(self):
        self.method = HttpMethod.OPTIONS
        return self.do_request()

    def __repr__(self):
        return "<IcvHttpClient> [{}] {}".format(self.method, self.url)



# -*- coding: UTF-8 -*-

import logging
from ..status_code import HTTP_STATUS_CODES

Log = logging.getLogger('simpleHttpServer.response')

class HttpResponse(object):

    def __init__(self, protocol, status_code, range=None):
        assert status_code in HTTP_STATUS_CODES, 'Unknown status code.'

        self.protocol = protocol
        self.status_code = status_code
        self.headers = {}
        self.range = range
        self.content = ''
        self.file = None

    def __str__(self):
        return 'HttpRequest (protocol=%s, status_code=%s)' % \
               (self.protocol, self.status_code)

    def write_to(self, output):
        if self.file:
            self.headers['Content-type'] = self.file.mime_type
            self.headers['Content-Length'] = self.file.file_size
            self.headers['Accept-Ranges'] = 'bytes'

            if self.range:
                range_start, range_end = self.file.calculate_range(self.range)

                self.headers['Content-Range'] = 'bytes %s-%s/%s' % (range_start, range_end,
                                                                    self.file.file_size)
                self.headers['Content-Length'] = range_end - range_start + 1

        response_msg = render_http_response(self)

        output.sendall(response_msg)

        Log.debug('Response:\n%s', response_msg)

        if self.file:
            self.file.stream_to(output, range=self.file.calculate_range(self.range))


def render_http_response(response):
    ret_val = []

    response_line = '%s %s %s' % (response.protocol, response.status_code,
                                  HTTP_STATUS_CODES[response.status_code][0])

    ret_val.append(response_line)

    for key, value in response.headers.iteritems():
        header_line = '%s: %s' % (key, value)
        ret_val.append(header_line)

    ret_val.append('')

    if response.content:
        ret_val.append(response.content)
    else:
        ret_val.append('')

    return '\n'.join(ret_val)


# -*- coding: UTF-8 -*-
from abc import ABCMeta, abstractclassmethod
from icv.utils import EasyDict


class Meta(object):
    __metaclass__ = ABCMeta


class SampleMeta(EasyDict):
    def dict(self):
        return super(SampleMeta, self).dict()


class AnnoMeta(EasyDict):
    def dict(self):
        return super(AnnoMeta, self).dict()



from .coco import Coco
from .voc import Voc
from .dataset import IcvDataSet
from .labelme import LabelMe
from .classify import Classify
from .shape import bbox_scaling, bbox_clip, bbox_extend
from .converter import VocConverter, CocoConverter, LabelMeConverter

__all__ = ['Coco', 'Voc', 'IcvDataSet', 'LabelMe', 'Classify', 'bbox_scaling', 'bbox_clip', 'bbox_extend',
           'VocConverter', 'CocoConverter', 'LabelMeConverter']



# -*- coding: UTF-8 -*-
try:
    import mmdet
except ModuleNotFoundError as e:
    raise Exception("You should install mmdetection first " \
                    "(https://github.com/open-mmlab/mmdetection/blob/master/INSTALL.md) . " \
                    "Import mmdet Error!")

MMDET_VERSION = int(mmdet.__version__.split("+")[0].replace(".", ""))

import os
import numpy as np
from ..data.core import BBox
from ..utils import ckpt_load, Config, Timer, is_file, concat_list
from ..image import imread, imwrite, imshow, imresize
from .detector import Detector
from .result import DetectionResult
import torch
from mmdet.models import build_detector
from mmdet.datasets import to_tensor
from mmdet.datasets.transforms import ImageTransform


class MmdetDetector(Detector):
    def __init__(self, model_path, config_file, categories, iou_thr=0.5, score_thr=0.5, device=None):
        super(MmdetDetector, self).__init__(categories=categories, iou_thr=iou_thr, score_thr=score_thr, device=device)

        self.model_path = model_path
        self.config_file = config_file
        self._build_detector()
        self.img_transform = ImageTransform(size_divisor=self.cfg.data.test.size_divisor, **self.cfg.img_norm_cfg)

    def _build_detector(self):
        self.model = self._init_detector(self.config_file, self.model_path, device=self.device)

    def _init_detector(self, config, checkpoint=None, device='cuda:0'):
        """Initialize a detector from config file.

        Args:
            config (str or :obj:`mmcv.Config`): Config file path or the config
                object.
            checkpoint (str, optional): Checkpoint path. If left as None, the model
                will not load any weights.

        Returns:
            nn.Module: The constructed detector.
        """
        if isinstance(config, str):
            self.cfg = Config.fromfile(config)
        elif not isinstance(config, Config):
            raise TypeError('config must be a filename or Config object, '
                            'but got {}'.format(type(config)))
        self.cfg.model.pretrained = None
        model = build_detector(self.cfg.model, test_cfg=self.cfg.test_cfg)
        try:
            if checkpoint is not None:
                checkpoint = ckpt_load(model, checkpoint)
            if 'meta' in checkpoint and 'CLASSES' in checkpoint['meta']:
                model.CLASSES = checkpoint['meta']['CLASSES']
            elif self.categories is not None:
                model.CLASSES = self.categories
            else:
                model.CLASSES = ["unknown"]
            model.cfg = self.cfg
        except:
            pass

        model.to(device)
        model.eval()
        return model

    def _prepare_data(self, img):
        ori_shape = img.shape
        img, img_shape, pad_shape, scale_factor = self.img_transform(
            img,
            scale=self.cfg.data.test.img_scale,
            keep_ratio=self.cfg.data.test.get('resize_keep_ratio', True))
        img = to_tensor(img).to(self.device).unsqueeze(0)
        img_meta = [
            dict(
                ori_shape=ori_shape,
                img_shape=img_shape,
                pad_shape=pad_shape,
                scale_factor=scale_factor,
                flip=False)
        ]
        return dict(img=[img], img_meta=[img_meta])

    def _inference_single(self, img):
        img = imread(img)
        data = self._prepare_data(img)
        with torch.no_grad():
            result = self.model(return_loss=False, rescale=True, **data)
        return result

    def _inference_batch(self, imgs):
        for img in imgs:
            yield self._inference_single(img)

    def _inference_detector(self, imgs):
        """Inference image(s) with the detector."""
        if not isinstance(imgs, list):
            return self._inference_single(imgs)
        else:
            return self._inference_batch(imgs)

    def _build_result(self, inference_result, inference_time=0, score_thr=-1):
        if isinstance(inference_result, tuple):
            bbox_result, segm_result = inference_result
        else:
            bbox_result, segm_result = inference_result, None

        detection_masks = []
        if segm_result is not None:
            import pycocotools.mask as maskUtils
            segms = concat_list(segm_result)
            for segm in segms:
                mask = maskUtils.decode(segm)
                detection_masks.append(mask)

        detection_masks = np.array(detection_masks)

        preds = [
            np.full(bbox.shape[0], i, dtype=np.int32)
            for i, bbox in enumerate(bbox_result)
        ]

        detection_classes = np.concatenate(preds)
        bbox_result = np.vstack(bbox_result)

        detection_bboxes = []
        detection_scores = []

        for ix in range(bbox_result.shape[0]):
            detection_bboxes.append(
                BBox(bbox_result[ix, 0], bbox_result[ix, 1], bbox_result[ix, 2], bbox_result[ix, 3]))
            detection_scores.append(bbox_result[ix, 4])

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result = DetectionResult(
            det_bboxes=detection_bboxes,
            det_classes=detection_classes + 1,
            det_scores=detection_scores,
            det_masks=detection_masks if detection_masks.shape[0] > 0 else None,  # TODO: process segm_result
            det_time=inference_time,
            categories=self.categories,
            score_thr=score_thr
        )

        return det_result

    def inference(self, image, is_show=False, save_path=None, score_thr=-1):
        image_np = imread(image)
        timer = Timer()
        result = self._inference_detector(image_np)
        inference_time = timer.since_start()

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result = self._build_result(result, inference_time, score_thr)
        det_result.vis(image_np)

        if is_show:
            imshow(det_result.det_image)

        if save_path is not None:
            imwrite(det_result.det_image, save_path)

        return det_result

    def inference_batch(self, image_batch, save_dir=None, resize=None, score_thr=-1):
        image_np_list = [imread(img) for img in image_batch]
        if resize:
            image_np_list = [imresize(img_np, resize) for img_np in image_np_list]

        timer = Timer()
        results = self._inference_detector(image_np_list)

        inference_time = timer.since_start()
        score_thr = score_thr if score_thr >= 0 else self.score_thr

        det_result_list = []
        for ix, result in enumerate(results):
            det_result = self._build_result(result, inference_time, score_thr)
            det_result.vis(image_np_list[ix])

            if save_dir is not None:
                save_path = os.path.join(save_dir, os.path.basename(image_batch[ix])) if is_file(
                    image_batch[0]) else os.path.join(save_dir, str(ix) + ".jpg")
                imwrite(det_result.det_image, save_path)

            det_result_list.append(det_result)

        return det_result_list



# transpose
FLIP_LEFT_RIGHT = 0
FLIP_TOP_BOTTOM = 1

from .bbox import BBox
from .bbox_list import BBoxList
from .mask import Mask
from .polys import Polygon
from .kps import Keypoint
from .image import Image
from .sample import Sample, Anno, AnnoShape
from .meta import SampleMeta, AnnoMeta



# -*- coding: UTF-8 -*-
import numpy as np

from .colorspace import bgr2rgb, rgb2bgr
from ..io import imread

def imnormalize(img, mean, std, to_rgb=True):
    img = imread(img).astype(np.float32)
    if to_rgb:
        img = bgr2rgb(img)
    return (img - mean) / std


def imdenormalize(img, mean, std, to_bgr=True):
    img = (imread(img) * std) + mean
    if to_bgr:
        img = rgb2bgr(img)
    return img



# -*- coding: UTF-8 -*-
from icv.utils import EasyDict as edict
from icv.image import imread, imdraw_bbox
from ..shape.transforms import bbox_clip, bbox_scaling, bbox_extend
import numpy as np


class BBox(object):
    def __init__(self, xmin, ymin, xmax, ymax, label=None, **kwargs):
        assert xmin <= xmax, "xmax should be large than xmin."
        assert ymin <= ymax, "ymax should be large than ymin."

        self.set_bbox(xmin, ymin, xmax, ymax)

        self.fields = edict()
        for k in kwargs:
            self.add_field(k, kwargs[k])

        self.label = label

        self.add_field("xmin", self.xmin)
        self.add_field("ymin", self.ymin)
        self.add_field("xmax", self.xmax)
        self.add_field("ymax", self.ymax)
        self.add_field("center", self.center)
        self.add_field("width", self.width)
        self.add_field("height", self.height)

    @staticmethod
    def init_from(bbox, label=None):
        assert isinstance(bbox, BBox) or isinstance(bbox, list) or isinstance(bbox, np.ndarray)
        if isinstance(bbox, np.ndarray):
            bbox = bbox.tolist()
        if isinstance(bbox, list):
            assert len(bbox) == 4
            bbox = BBox(xmin=bbox[0], ymin=bbox[1], xmax=bbox[2], ymax=bbox[3], label=label)
        return bbox

    def set_bbox(self, xmin, ymin, xmax, ymax):
        self._xmin = xmin
        self._ymin = ymin
        self._xmax = xmax
        self._ymax = ymax

        self._center = ((xmax + xmin) / 2, (ymax + ymin) / 2)
        self._width = xmax - xmin
        self._height = ymax - ymin

    def add_field(self, field, field_data):
        self.fields[field] = field_data

    def get_field(self, field):
        return self.fields[field]

    def has_field(self, field):
        return field in self.fields

    def fields(self):
        return list(self.fields.keys())

    @property
    def bbox(self):
        return [self._xmin, self._ymin, self._xmax, self._ymax]

    @property
    def xmin(self):
        return self._xmin

    @property
    def ymin(self):
        return self._ymin

    @property
    def xmax(self):
        return self._xmax

    @property
    def ymax(self):
        return self._ymax

    @property
    def xmin_int(self):
        return int(np.round(self._xmin))

    @property
    def ymin_int(self):
        return int(np.round(self._ymin))

    @property
    def xmax_int(self):
        return int(np.round(self._xmax))

    @property
    def ymax_int(self):
        return int(np.round(self._ymax))

    @property
    def center(self):
        return self._center

    @property
    def center_x(self):
        return self.xmin + self._width / 2

    @property
    def center_y(self):
        return self.ymin + self._height / 2

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def ratio(self):
        return round(self._height / self._width, 2)

    @property
    def area(self):
        return self._height * self._width

    def iou(self, other):
        bbox = BBox.init_from(other)
        # TODO: check it
        # determine the (x, y)-coordinates of the intersection rectangle
        xA = max(self.xmin, bbox.xmin)
        yA = max(self.ymin, bbox.ymin)
        xB = min(self.xmax, bbox.xmax)
        yB = min(self.ymax, bbox.ymax)

        # compute the area of intersection rectangle
        interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)

        # compute the area of both the prediction and ground-truth
        # rectangles
        boxAArea = (self.xmax - self.xmin + 1) * (self.ymax - self.ymin + 1)
        boxBArea = (bbox.xmax - bbox.xmin + 1) * (bbox.ymax - bbox.ymin + 1)

        # compute the intersection over union by taking the intersection
        # area and dividing it by the sum of prediction + ground-truth
        # areas - the interesection area
        iou = interArea / float(boxAArea + boxBArea - interArea)

        # return the intersection over union value
        return iou

    def intersection(self, other, default=None):
        bbox = BBox.init_from(other)
        xA = max(self.xmin, bbox.xmin)
        yA = max(self.ymin, bbox.ymin)
        xB = min(self.xmax, bbox.xmax)
        yB = min(self.ymax, bbox.ymax)

        if xA > xB or yA > yB:
            return default
        else:
            return BBox(xA, yA, xB, yB)

    def union(self, other):
        bbox = BBox.init_from(other)
        return BBox(
            xmin=min(self.xmin, bbox.xmin),
            ymin=min(self.ymin, bbox.ymin),
            xmax=max(self.xmax, bbox.xmax),
            ymax=max(self.ymax, bbox.ymax)
        )

    def is_fully_within_image(self, image):
        """
        Estimate whether the bounding box is fully inside the image area.
        Parameters
        ----------
        image : (H,W,...) ndarray or tuple of int
            Image dimensions to use.
            If an ndarray, its shape will be used.
            If a tuple, it is assumed to represent the image shape
            and must contain at least two integers.
        Returns
        -------
        bool
            True if the bounding box is fully inside the image area. False otherwise.
        """
        shape = imread(image).shape
        height, width = shape[0:2]
        return self.xmin >= 0 and self.xmax < width and self.ymin >= 0 and self.ymax < height

    def is_partly_within_image(self, image):
        """
        Estimate whether the bounding box is at least partially inside the image area.
        Parameters
        ----------
        image : (H,W,...) ndarray or tuple of int
            Image dimensions to use.
            If an ndarray, its shape will be used.
            If a tuple, it is assumed to represent the image shape
            and must contain at least two integers.
        Returns
        -------
        bool
            True if the bounding box is at least partially inside the image area. False otherwise.
        """
        shape = imread(image).shape
        height, width = shape[0:2]
        eps = np.finfo(np.float32).eps
        img_bb = BBox(xmin=0, xmax=width - eps, ymin=0, ymax=height - eps)
        return self.intersection(img_bb) is not None

    def is_out_of_image(self, image, fully=True, partly=False):
        """
        Estimate whether the bounding box is partially or fully outside of the image area.
        Parameters
        ----------
        image : (H,W,...) ndarray or tuple of int
            Image dimensions to use. If an ndarray, its shape will be used. If a tuple, it is
            assumed to represent the image shape and must contain at least two integers.
        fully : bool, optional
            Whether to return True if the bounding box is fully outside fo the image area.
        partly : bool, optional
            Whether to return True if the bounding box is at least partially outside fo the
            image area.
        Returns
        -------
        bool
            True if the bounding box is partially/fully outside of the image area, depending
            on defined parameters. False otherwise.
        """
        if self.is_fully_within_image(image):
            return False
        elif self.is_partly_within_image(image):
            return partly
        else:
            return fully

    def copy(self, xmin=None, ymin=None, xmax=None, ymax=None, label=None, **kwargs):
        """
        Create a shallow copy of the BoundingBox object.
        Parameters
        ----------
        xmin : None or number
            If not None, then the xmin coordinate of the copied object will be set to this value.
        ymin : None or number
            If not None, then the ymin coordinate of the copied object will be set to this value.
        xmax : None or number
            If not None, then the xmax coordinate of the copied object will be set to this value.
        ymax : None or number
            If not None, then the ymax coordinate of the copied object will be set to this value.
        label : None or string
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.BoundingBox
            Shallow copy.
        """
        return BBox(
            xmin=self.xmin if xmin is None else xmin,
            xmax=self.xmax if xmax is None else xmax,
            ymin=self.ymin if ymin is None else ymin,
            ymax=self.ymax if ymax is None else ymax,
            label=self.label if label is None else label,
            **kwargs
        )

    def deepcopy(self, xmin=None, ymin=None, xmax=None, ymax=None, label=None, **kwargs):
        """
        Create a deep copy of the BoundingBox object.
        Parameters
        ----------
        xmin : None or number
            If not None, then the xmin coordinate of the copied object will be set to this value.
        ymin : None or number
            If not None, then the ymin coordinate of the copied object will be set to this value.
        xmax : None or number
            If not None, then the xmax coordinate of the copied object will be set to this value.
        ymax : None or number
            If not None, then the ymax coordinate of the copied object will be set to this value.
        label : None or string
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.BoundingBox
            Deep copy.
        """
        return self.copy(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, label=label, **kwargs)

    def draw_on_image(self, image, color=(0, 255, 0), thickness=1,
                      copy=True, raise_if_out_of_image=False):
        """
        Draw the bounding box on an image.
        Parameters
        ----------
        image : (H,W,C) ndarray(uint8)
            The image onto which to draw the bounding box.
        color : iterable of int, optional
            The color to use, corresponding to the channel layout of the image. Usually RGB.
        alpha : float, optional
            The transparency of the drawn bounding box, where 1.0 denotes no transparency and
            0.0 is invisible.
        size : int, optional
            The thickness of the bounding box in pixels. If the value is larger than 1, then
            additional pixels will be added around the bounding box (i.e. extension towards the
            outside).
        copy : bool, optional
            Whether to copy the input image or change it in-place.
        raise_if_out_of_image : bool, optional
            Whether to raise an error if the bounding box is fully outside of the
            image. If set to False, no error will be raised and only the parts inside the image
            will be drawn.
        Returns
        -------
        result : (H,W,C) ndarray(uint8)
            Image with bounding box drawn on it.
        """
        # from icv.image.vis import imdraw_bbox

        if raise_if_out_of_image and self.is_out_of_image(image):
            raise Exception(
                "Cannot draw bounding box xmin=%.8f, ymin=%.8f, xmax=%.8f, ymax=%.8f on image with shape %s." % (
                    self.xmin, self.ymin, self.xmax, self.ymax, image.shape))

        result = np.copy(image) if copy else image

        display = "%s %.2f" % (
            self.label if self.label is not None else "",
            self.get_field("score") if self.has_field("score") else 0.0
        )

        return imdraw_bbox(
            result,
            xmin=self._xmin,
            ymin=self._ymin,
            xmax=self._xmax,
            ymax=self._ymax,
            color=color,
            thickness=thickness,
            display_str=display
        )

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        s = self.__class__.__name__ + "("
        s += "xmin={}, ".format(self.xmin)
        s += "ymin={}, ".format(self.ymin)
        s += "xmax={}, ".format(self.xmax)
        s += "ymax={}, ".format(self.ymax)
        return s

    def clip(self, img_shape):
        clipped_bbox = bbox_clip(self, img_shape)[0]
        self.set_bbox(clipped_bbox[0], clipped_bbox[1], clipped_bbox[2], clipped_bbox[3])

    def scaling(self, scale, clip_shape=None):
        scaled_bbox = bbox_scaling(self, scale, clip_shape)[0]
        self.set_bbox(scaled_bbox[0], scaled_bbox[1], scaled_bbox[2], scaled_bbox[3])

    def extend(self, pad, clip_shape=None):
        extended_bbox = bbox_extend(self, pad, clip_shape)[0]
        self.set_bbox(extended_bbox[0], extended_bbox[1], extended_bbox[2], extended_bbox[3])



from .io import imread, imread_topil, imread_tob64, imwrite, np_img_to_pil, pil_img_to_np
from .vis import imshow, imshow_bboxes, imdraw_bbox, imdraw_mask, imdraw_polygons, imdraw_polygons_with_bbox,imdraw_text
from .transforms import *

__all__ = [
    'imread', 'imread_topil', 'imread_tob64', 'imwrite', 'np_img_to_pil', 'pil_img_to_np',
    'imshow', 'imshow_bboxes', 'imdraw_bbox', 'imdraw_mask', 'imdraw_polygons', 'imdraw_polygons_with_bbox','imdraw_text',
    'imresize', 'imresize_like', 'imrescale', 'bgr2gray', 'gray2bgr', 'bgr2rgb', 'rgb2bgr', 'bgr2hsv', 'hsv2bgr',
    'immerge','immix_up','imflip_lr', 'imflip_ud', 'imrotate', 'imcrop', 'impad', 'impad_to_multiple', 'imnormalize', 'imdenormalize'
]



# -*- coding: UTF-8 -*-
import json
from .itis import is_file

def encode_to_file(obj,filepath,overwrite=True):
    """
    save obj to json file, like json.dump()
    :param filepath:
    :param obj:
    :return:
    """
    if not overwrite and is_file(filepath):
        return
    return json.dump(obj,open(filepath,"w"))

def decode_from_file(filepath):
    """
    parse json file to obj, like json.load()
    :param filepath:
    :return:
    """
    return json.load(open(filepath))

def json_encode(obj):
    """
    encode obj to json string, like json.dumps()
    :param obj:
    :return:
    """
    return json.dumps(obj)


def json_decode(str, encoding="utf-8"):
    """
    decode json string to json object, like json.loads()
    :param str:
    :param encoding:
    :return:
    """
    return json.loads(str,encoding=encoding)



# -*- coding: UTF-8 -*-
import shutil
import os
import six
from .itis import is_seq,is_file,is_dir

def mkfile(filepath,content=""):
    if not is_dir(os.path.dirname(filepath)):
        mkdir(os.path.dirname(filepath))
    with open(filepath,"w") as f:
        f.write(content)

def mkdir(dir_name,mode=0o777):
    if six.PY3:
        os.makedirs(dir_name, mode=mode, exist_ok=True)
    else:
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name, mode=mode)

def mkdir_force(dir_name, subdirs=None,mode=0o777):
    assert subdirs is None or is_seq(subdirs),"param subdirs should be none or a sequence"
    if os.path.exists(dir_name):
        shutil.rmtree(dir_name)
    mkdir(dir_name,mode)

    if subdirs is not None:
        for subdir in subdirs:
            mkdir(os.path.join(dir_name,subdir),mode)

def reset_dir(dir_name):
    if os.path.exists(dir_name):
        shutil.rmtree(dir_name)
    os.makedirs(dir_name)

def fcopy(source_path,dist_path,rename=True):
    """拷贝文件或文件列表到目标目录中
    :param source_path: 源文件或源文件列表
    :param dist_path: 目标目录或目标文件
    :param rename: 是否重命名
    :return: 目标文件路径
    """
    if is_seq(source_path):
        source_path = list(source_path)
        dist_file_list = []
        for source_file in source_path:
            dist_file_list.extend(fcopy(source_file, dist_path, rename))
        return dist_file_list

    assert not (is_dir(source_path) and is_file(dist_path)), "Can't copy dir to file"
    if not is_file(source_path) and not is_dir(source_path):
        return None
    if is_file(source_path):
        dist_path = os.path.join(dist_path,os.path.basename(source_path)) if is_dir(dist_path) else dist_path
        i = 0
        while is_file(dist_path) and rename:
            i += 1
            dist_name,dist_ext = os.path.basename(dist_path).rsplit(".",1)
            dist_path = os.path.join(os.path.dirname(dist_path),"%s_%d.%s" % (dist_name,i,dist_ext))

        shutil.copyfile(source_path,dist_path)
        return [dist_path]
    else:
        dist_file_list = []
        for f in os.listdir(source_path):
            source_file = os.path.join(source_path,f)
            dist_file_list.extend(fcopy(source_file,dist_path,rename))

        return dist_file_list

def list_from_file(filename, prefix='', offset=0, max_num=0):
    """Load a text file and parse the content as a list of strings.

    Args:
        filename (str): Filename.
        prefix (str): The prefix to be inserted to the begining of each item.
        offset (int): The offset of lines.
        max_num (int): The maximum number of lines to be read,
            zeros and negatives mean no limitation.

    Returns:
        list[str]: A list of strings.
    """
    cnt = 0
    item_list = []
    with open(filename, 'r') as f:
        for _ in range(offset):
            f.readline()
        for line in f:
            if max_num > 0 and cnt >= max_num:
                break
            item_list.append(prefix + line.rstrip('\n'))
            cnt += 1
    return item_list

def list_to_file(seq, filename, prefix=''):
    seq = [prefix + str(s) for s in seq]
    with open(filename,"w") as f:
        f.write("\n".join(seq))
    return filename


from .show import *
from .color import STANDARD_COLORS, MASK_COLORS, DARK_COLORS, LIGHT_COLORS, get_color_tuple,get_reverse_color
from .plot import draw_line, draw_bar

__all__ = ['STANDARD_COLORS', 'MASK_COLORS', 'DARK_COLORS', 'LIGHT_COLORS', 'get_color_tuple', 'get_reverse_color','draw_line', 'draw_bar']



# -*- coding: UTF-8 -*-
import numpy as np
import matplotlib.pyplot as plt
from ..utils.itis import is_str, is_seq

# set draw style
# plt.style.use("ggplot")
# for Chinese
# plt.rcParams["font.sans-serif"] = ["FangSong"]
plt.rcParams["axes.unicode_minus"] = False

plt.tick_params(top='off', right='off')
marks = ['o', 'x']

def draw_line(
        y_data,
        x_data=None,
        figsize=None,
        show_value=False,
        legends=None,
        title=None,
        x_label=None,
        y_label=None,
        x_ticklabels=None,
        y_ticklabels=None,
        x_tick_rotation=0,
        y_tick_rotation=0,
        x_line_values=None,
        y_line_values=None,
        marker='o',
        linestyle='-',
        linewidth=2,
        color='r',
        save_path=None,
):
    y_data = np.array(y_data)
    assert y_data.ndim <= 2
    cnt = y_data.shape[0]
    if x_data is not None:
        x_data = np.array(x_data)
        assert x_data.shape == y_data.shape
    else:
        x_data = np.array([i for i in range(y_data.shape[1])] * cnt).reshape(y_data.shape)

    if y_data.ndim == 1:
        y_data = y_data[None, ::]
        x_data = x_data[None, ::]

    assert figsize is None or (isinstance(figsize, tuple) and len(figsize) == 2)
    assert legends is None or (is_seq(legends) and len(legends) == cnt)

    if is_str(color):
        color = [color] * cnt
    elif is_seq(color):
        color = list(color)[:cnt]
        if len(color) < cnt:
            color = color + [color[-1]] * (cnt - len(color))
    else:
        color = ["r"] * cnt

    for i in range(cnt):
        marker = marker if marker is not None else marks[i % len(marks)]
        plt.plot(
            x_data[i, :], y_data[i, :], linestyle=linestyle, linewidth=linewidth, color=color[i],
            marker=marker,
            label=None if legends is None else legends[i]
        )

        if show_value:
            for a, b in zip(x_data[i, :], y_data[i, :]):
                plt.annotate('%.3f' % b, xy=(a, b), xytext=(-20, 10),
                             textcoords='offset points')

    if title is not None:
        plt.title(title)

    if x_label is not None:
        plt.xlabel(x_label)

    if y_label is not None:
        plt.ylabel(y_label)

    if x_ticklabels is not None:
        plt.gca().set_xticklabels(x_ticklabels)
        if x_tick_rotation > 0:
            for tick in plt.gca().get_xticklabels():
                tick.set_rotation(x_tick_rotation)

    if y_ticklabels is not None:
        plt.gca().set_yticklabels(y_ticklabels)
        if y_tick_rotation > 0:
            for tick in plt.gca().get_yticklabels():
                tick.set_rotation(y_tick_rotation)

    if x_line_values is not None:
        if not is_seq(x_line_values):
            x_line_values = [x_line_values]

        for x in x_line_values:
            plt.axvline(x=x,ymax=np.max(y_data),linestyle="--")

    if y_line_values is not None:
        if not is_seq(y_line_values):
            y_line_values = [y_line_values]

        for y in y_line_values:
            plt.axhline(y=y,linestyle="--")

    if legends is not None:
        plt.legend()

    if save_path is not None:
        plt.savefig(save_path)

    plt.show()


def draw_bar(
        data_list,
        show_value=True,
        legends=None,
        title=None,
        x_ticklabels=None,
        y_ticklabels=None,
        x_tick_rotation=0,
        y_tick_rotation=0,
        alpha=0.9,
        save_path=None,
):
    assert len(data_list) > 0 and is_seq(data_list[0])
    assert legends is None or (is_seq(legends) and len(legends) == len(data_list))

    width = 1 / len(data_list)
    for ix, data in enumerate(data_list):
        plt.bar(list(range(len(data))), data, width=width, label=legends[ix], alpha=alpha, tick_label=x_ticklabels)
        if show_value:
            for i in range(len(data)):
                plt.text(i, data[i], "%.3f" % data[i], ha="center", va="bottom", fontsize=11)

    if x_ticklabels is not None:
        if x_tick_rotation > 0:
            for tick in plt.gca().get_xticklabels():
                tick.set_rotation(x_tick_rotation)

    if y_ticklabels is not None:
        plt.gca().set_yticklabels(y_ticklabels)
        if y_tick_rotation > 0:
            for tick in plt.gca().get_yticklabels():
                tick.set_rotation(y_tick_rotation)

    if legends is not None:
        plt.legend()

    if title is not None:
        plt.title(title)

    if save_path is not None:
        plt.savefig(save_path)

    plt.show()



# -*- coding: UTF-8 -*-
import numpy as np
from .bbox import BBox
from .polys import Polygon
from .mask import Mask


class AnnoShape(object):
    def __init__(self, bbox=None, polys=None, mask=None, label=None):
        self.bbox = None if bbox is None else BBox.init_from(bbox, label)
        self.polys = None if polys is None else Polygon.init_from(polys, label)
        self.mask = None if mask is None else Mask.init_from(mask, label)

        self.label = label

        assert not (self.seg_mode_polys and self.seg_mode_mask)

        if self.bbox is None and self.seg_mode_polys:
            self.bbox = self.polys.to_bounding_box()

        if self.bbox is None and self.seg_mode_mask:
            self.bbox = self.mask.to_bounding_box()

    @property
    def seg_mode_polys(self):
        return self.polys is not None

    @property
    def seg_mode_mask(self):
        return self.mask is not None

    @property
    def seg_mode(self):
        return self.seg_mode_mask or self.seg_mode_polys



# -*- coding: UTF-8 -*-
import numpy as np
from .kps import Keypoint
from icv.utils import EasyDict as edict
from icv.utils import is_np_array
from icv.image.vis import imdraw_polygons

class Polygon(object):
    def __init__(self, exterior, label=None, **kwargs):
        if isinstance(exterior, list):
            if not exterior:
                # for empty lists, make sure that the shape is (0, 2) and not (0,) as that is also expected when the
                # input is a numpy array
                self.exterior = np.zeros((0, 2), dtype=np.float32)
            elif isinstance(exterior[0], Keypoint):
                # list of Keypoint
                self.exterior = np.float32([[point.x, point.y] for point in exterior])
            elif not isinstance(exterior[0], (list, tuple)):
                assert len(exterior) % 2 == 0
                self.exterior = []
                for i, v in enumerate(exterior):
                    if i % 2 == 0:
                        self.exterior.append([v, exterior[i + 1]])
                self.exterior = np.float32(self.exterior)
            else:
                # list of tuples (x, y)
                self.exterior = np.float32([[point[0], point[1]] for point in exterior])
        else:
            assert is_np_array(exterior), ("Expected exterior to be a list of tuples (x, y) or "
                                           + "an (N, 2) array, got type %s") % (exterior,)

            assert exterior.ndim == 2 and exterior.shape[1] == 2, ("Expected exterior to be a list of tuples (x, y) or "
                                                                   "an (N, 2) array, got an array of shape %s") % (
                                                                      exterior.shape,)

            self.exterior = np.float32(exterior)

        if len(self.exterior) >= 2 and np.allclose(self.exterior[0, :], self.exterior[-1, :]):
            self.exterior = self.exterior[:-1]

        self.label = label

        self.bbox = None
        self.mask = None

        self.fields = edict()
        for k in kwargs:
            self.add_field(k, kwargs[k])

        self.label = label

        self.add_field("xx", self.xx)
        self.add_field("xx_int", self.xx_int)
        self.add_field("yy", self.yy)
        self.add_field("yy_int", self.yy_int)
        self.add_field("is_valid", self.is_valid)
        self.add_field("area", self.area)
        self.add_field("height", self.height)
        self.add_field("width", self.width)

    @staticmethod
    def init_from(polys, label=None):
        if isinstance(polys, list) or is_np_array(polys):
            return Polygon(polys, label)
        elif isinstance(polys, Polygon):
            return polys.deepcopy(label=label)
        else:
            raise TypeError("Type of polys should be list or ndarray or Polygon")

    def add_field(self, field, field_data):
        self.fields[field] = field_data

    def get_field(self, field):
        return self.fields[field]

    def has_field(self, field):
        return field in self.fields

    def fields(self):
        return list(self.fields.keys())

    @property
    def xx(self):
        """
        Return the x-coordinates of all points in the exterior.
        Returns
        -------
        (N,2) ndarray
            X-coordinates of all points in the exterior as a float32 ndarray.
        """
        return self.exterior[:, 0]

    @property
    def yy(self):
        """
        Return the y-coordinates of all points in the exterior.
        Returns
        -------
        (N,2) ndarray
            Y-coordinates of all points in the exterior as a float32 ndarray.
        """
        return self.exterior[:, 1]

    @property
    def xx_int(self):
        """
        Return the x-coordinates of all points in the exterior, rounded to the closest integer value.
        Returns
        -------
        (N,2) ndarray
            X-coordinates of all points in the exterior, rounded to the closest integer value.
            Result dtype is int32.
        """
        return np.int32(np.round(self.xx))

    @property
    def yy_int(self):
        """
        Return the y-coordinates of all points in the exterior, rounded to the closest integer value.
        Returns
        -------
        (N,2) ndarray
            Y-coordinates of all points in the exterior, rounded to the closest integer value.
            Result dtype is int32.
        """
        return np.int32(np.round(self.yy))

    @property
    def is_valid(self):
        """
        Estimate whether the polygon has a valid shape.
        To to be considered valid, the polygons must be made up of at least 3 points and have concave shape.
        Multiple consecutive points are allowed to have the same coordinates.
        Returns
        -------
        bool
            True if polygon has at least 3 points and is concave, otherwise False.
        """
        if len(self.exterior) < 3:
            return False
        return self.to_shapely_polygon().is_valid

    @property
    def area(self):
        """
        Estimate the area of the polygon.
        Returns
        -------
        number
            Area of the polygon.
        """
        if len(self.exterior) < 3:
            raise Exception("Cannot compute the polygon's area because it contains less than three points.")
        poly = self.to_shapely_polygon()
        return poly.area

    @property
    def height(self):
        """
        Estimate the height of the polygon.
        Returns
        -------
        number
            Height of the polygon.
        """
        yy = self.yy
        return max(yy) - min(yy)

    @property
    def width(self):
        """
        Estimate the width of the polygon.
        Returns
        -------
        number
            Width of the polygon.
        """
        xx = self.xx
        return max(xx) - min(xx)

    @property
    def polygons(self):
        return np.array(self.exterior).reshape(-1)

    def to_shapely_polygon(self):
        """
        Convert this polygon to a Shapely polygon.
        Returns
        -------
        shapely.geometry.Polygon
            The Shapely polygon matching this polygon's exterior.
        """
        # load shapely lazily, which makes the dependency more optional
        import shapely.geometry
        return shapely.geometry.Polygon([(point[0], point[1]) for point in self.exterior])

    def to_bounding_box(self):
        """
        Convert this polygon to a bounding box tightly containing the whole polygon.
        Returns
        -------
        imgaug.BoundingBox
            Tight bounding box around the polygon.
        """
        if self.bbox is not None:
            return self.bbox
        from .bbox import BBox

        xx = self.xx
        yy = self.yy
        self.bbox = BBox(xmin=min(xx), xmax=max(xx), ymin=min(yy), ymax=max(yy), label=self.label, **self.fields)
        return self.bbox

    def to_keypoints(self):
        """
        Convert this polygon's `exterior` to ``Keypoint`` instances.
        Returns
        -------
        list of imgaug.Keypoint
            Exterior vertices as ``Keypoint`` instances.
        """
        from .kps import Keypoint

        return [Keypoint(x=point[0], y=point[1]) for point in self.exterior]

    def to_mask(self, height, width):
        if self.mask is not None:
            return self.mask

        import pycocotools.mask as mask_utils
        from .mask import Mask
        rles = mask_utils.frPyObjects([self.polygons], height, width)
        rle = mask_utils.merge(rles)
        mask = mask_utils.decode(rle)
        self.mask = Mask(mask)
        return self.mask

    @staticmethod
    def from_shapely(polygon_shapely, label=None):
        """
        Create a polygon from a Shapely polygon.
        Note: This will remove any holes in the Shapely polygon.
        Parameters
        ----------
        polygon_shapely : shapely.geometry.Polygon
             The shapely polygon.
        label : None or str, optional
            The label of the new polygon.
        Returns
        -------
        imgaug.Polygon
            A polygon with the same exterior as the Shapely polygon.
        """
        # load shapely lazily, which makes the dependency more optional
        import shapely.geometry

        assert isinstance(polygon_shapely, shapely.geometry.Polygon)

        # polygon_shapely.exterior can be None if the polygon was instantiated without points
        if polygon_shapely.exterior is None or len(polygon_shapely.exterior.coords) == 0:
            return Polygon([], label=label)
        exterior = np.float32([[x, y] for (x, y) in polygon_shapely.exterior.coords])
        return Polygon(exterior, label=label)

    def copy(self, exterior=None, label=None, **kwargs):
        """
        Create a shallow copy of the Polygon object.
        Parameters
        ----------
        exterior : list of imgaug.Keypoint or list of tuple or (N,2) ndarray, optional
            List of points defining the polygon. See :func:`imgaug.Polygon.__init__` for details.
        label : None or str, optional
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.Polygon
            Shallow copy.
        """
        return self.deepcopy(exterior=exterior, label=label, **kwargs)

    def deepcopy(self, exterior=None, label=None, **kwargs):
        """
        Create a deep copy of the Polygon object.
        Parameters
        ----------
        exterior : list of Keypoint or list of tuple or (N,2) ndarray, optional
            List of points defining the polygon. See `imgaug.Polygon.__init__` for details.
        label : None or str
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.Polygon
            Deep copy.
        """
        return Polygon(
            exterior=np.copy(self.exterior) if exterior is None else exterior,
            label=self.label if label is None else label,
            **kwargs
        )

    def draw_on_image(self,
                      image,
                      color=(0, 255, 0), alpha=0.6,
                      copy=True, raise_if_out_of_image=False):

        bbox = self.to_bounding_box()
        if raise_if_out_of_image and bbox.is_out_of_image(image):
            raise Exception(
                "Cannot draw mask xmin = %.8f, ymin = %.8f, xmax = %.8f, ymax = %.8f on image with shape %s." % (
                    bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax, image.shape))

        result = np.copy(image) if copy else image

        result = imdraw_polygons(result, [self.exterior.tolist()], color=color, alpha=alpha)
        return result

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        points_str = ", ".join(["(x=%.3f, y=%.3f)" % (point[0], point[1]) for point in self.exterior])
        return "Polygon([%s] (%d points), label=%s)" % (points_str, len(self.exterior), self.label)



# -*- coding: UTF-8 -*-
from ..utils.itis import is_seq, is_str

STANDARD_COLORS = [
    'AliceBlue', 'Chartreuse', 'Aqua', 'Aquamarine', 'Azure', 'Beige', 'Bisque',
    'BlanchedAlmond', 'BlueViolet', 'BurlyWood', 'CadetBlue', 'AntiqueWhite',
    'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson', 'Cyan',
    'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
    'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
    'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
    'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'Gold', 'GoldenRod',
    'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
    'Lavender', 'LavenderBlush', 'LawnGreen', 'LemonChiffon', 'LightBlue',
    'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
    'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
    'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
    'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
    'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
    'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
    'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
    'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
    'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
    'Red', 'RosyBrown', 'RoyalBlue', 'SaddleBrown', 'Green', 'SandyBrown',
    'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
    'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
    'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
    'WhiteSmoke', 'Yellow', 'YellowGreen'
]

# DARK_COLORS = [
#     (255, 48, 48), (148, 0, 211), (160, 82, 45), (139, 0, 0), (71, 71, 71), (34, 139, 34), (3, 3, 3), (0, 0, 238),
#     (238, 0, 238)
# ]

DARK_COLORS = ["darkblue", "darkcyan", "darkgoldenrod", "darkgray", "darkgrey", "darkgreen",
               "darkkhaki", "darkmagenta", "darkolivegreen", "darkorange", "darkorchid", "darkred",
               "darksalmon", "darkseagreen", "darkslateblue", "darkslategray", "darkslategrey",
               "darkturquoise", "darkviolet"]

LIGHT_COLORS = ['lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrodyellow', 'lightgreen',
                'lightgray', 'lightgrey', 'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue',
                'lightslategray', 'lightslategrey', 'lightsteelblue', 'lightyellow']

MASK_COLORS = [
    (0, 0, 139),
    (67, 205, 128),
    (0, 206, 209),
    (255, 106, 106),
    (192, 255, 62),
    (130, 130, 130),
    (139, 87, 66),
    (25, 25, 112)
]


def get_color_tuple(color, alpha=None):
    if alpha is not None:
        alpha = min(255, int(alpha)) if alpha > 1 else int(alpha * 255)
    if isinstance(color, (int, float)):
        if color < 1:
            color = int(255 * color)
        color = min(255, int(color))
        if alpha is not None:
            return (color, color, color, alpha)
        return (color, color, color)
    elif is_seq(color):
        color = tuple(color)[:4]
        if alpha is None:
            return color
        if len(color) == 3:
            color = color + (alpha,)
        elif len(color) == 4:
            color = color[:3] + (alpha)
        else:
            raise ValueError("color value invalid")
        return color
    elif is_str(color):
        from PIL import ImageColor
        return ImageColor.getrgb(color)
    else:
        raise ValueError("color value invalid")


def get_reverse_color(color):
    c = get_color_tuple(color)
    reverse_color = []
    for i in c:
        reverse_color.append(255 - i)
    return tuple(reverse_color)




# -*- coding: UTF-8 -*-
from .dataset import IcvDataSet
from ..utils import is_seq, is_dir
from ..image import imwrite
from ..data.core.bbox import BBox
from ..data.core.polys import Polygon
from ..data.core.sample import Sample, Anno
from ..data.core.meta import AnnoMeta,SampleMeta
from ..vis.color import STANDARD_COLORS
import random
import os
import json
import shutil
from tqdm import tqdm


class LabelMe(IcvDataSet):
    def __init__(self, image_anno_path_list, split="trainval", keep_no_anno_image=True, categories=None,
                 one_index=False):
        assert is_seq(image_anno_path_list)
        image_anno_path_list = list(image_anno_path_list)
        image_path_list, anno_path_list = list(zip(*image_anno_path_list))

        self.split = split
        self.keep_no_anno_image = keep_no_anno_image
        self.one_index = one_index

        self.ids = [os.path.basename(_).rsplit(".", 1)[0] for _ in image_path_list]
        self.id2imgpath = {id: image_path_list[ix] for ix, id in enumerate(self.ids)}
        self.id2annopath = {id: anno_path_list[ix] for ix, id in enumerate(self.ids)}

        self.sample_db = {}
        self.color_map = {}
        self.get_samples()
        self.categories = categories if categories is not None else self.get_categories()

        super(LabelMe, self).__init__(self.ids, self.categories, self.keep_no_anno_image, one_index)
        print("there have %d samples in LabelMe dataset" % len(self.ids))
        print("there have %d categories in LabelMe dataset" % len(self.categories))

    @property
    def is_seg_mode(self):
        if self.length == 0:
            return False

        for i in range(self.length):
            for anno in self.get_sample(self.ids[i]).annos:
                if anno.seg_mode_polys or anno.seg_mode_mask:
                    return True

        return False

    def get_categories(self):
        categories = []
        for anno_sample in self.samples:
            label_list = [anno.label for anno in anno_sample.annos]
            if label_list:
                categories.extend(label_list)
        categories = list(set(categories))
        categories.sort()
        self.set_categories(categories)
        return categories

    def save(self, output_dir, reset_dir=False, split=None):
        anno_path, image_path = LabelMe.reset_dir(output_dir, reset=reset_dir)
        for id in self.ids:
            self._write(self.get_sample(id), anno_path, image_path)

    @staticmethod
    def reset_dir(dist_dir, reset=False):
        if not reset:
            assert is_dir(dist_dir)
        if reset and os.path.exists(dist_dir):
            shutil.rmtree(dist_dir)

        anno_path = os.path.join(dist_dir, "annotations")
        image_path = os.path.join(dist_dir, "images")

        for _path in [anno_path, image_path]:
            if reset or not is_dir(_path):
                os.makedirs(_path)

        return anno_path, image_path

    def _get_bbox_from_points(self, points):
        """
        根据polygon顶点获取bbox
        :param points:
        :return:
        """
        x_list = [p[0] for p in points]
        y_list = [p[1] for p in points]

        xmin = min(x_list)
        ymin = min(y_list)
        xmax = max(x_list)
        ymax = max(y_list)

        return xmin, ymin, xmax, ymax

    def get_sample(self, id):
        """
        get sample
        :param id: image name
        :return:
        """
        if id in self.sample_db:
            return self.sample_db[id]

        anno_file = self.id2annopath[id]
        anno_data = json.load(open(anno_file, "r"))
        img_file = self.id2imgpath[id]

        annos = []
        if "shapes" in anno_data:
            shapes = anno_data["shapes"]
            for shape in shapes:
                if "shape_type" not in shape or "points" not in shape or "label" not in shape:
                    continue

                points = shape["points"]
                xmin, ymin, xmax, ymax = self._get_bbox_from_points(points)

                label = shape["label"]
                if label not in self.color_map:
                    self.color_map[label] = random.choice(STANDARD_COLORS)

                anno = Anno(
                    bbox=BBox(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, label=label),
                    label=label,
                    color=self.color_map[label],
                    polys=Polygon.init_from(points, label=label) if shape["shape_type"] == "polygon" else None,
                    meta=AnnoMeta()

                )
                annos.append(anno)

        sample = Sample(
            name=id,
            image=img_file,
            annos=annos,
            meta=SampleMeta()
        )
        return sample

    def _write(self, anno_sample, anno_path, img_path):
        assert isinstance(anno_sample, Sample)

        if is_dir(anno_path):
            anno_path = os.path.join(anno_path, "%s.json" % anno_sample.name)

        if is_dir(img_path):
            img_path = os.path.join(img_path, "%s.jpg" % anno_sample.name)

        imwrite(anno_sample.image, img_path)
        anno_json = {
            "shapes": [],
            "imagePath": img_path,
            "imageHeight": anno_sample.height,
            "imageWidth": anno_sample.width
        }

        for anno in anno_sample.annos:
            shape = {
                "label": anno.label,
                "shape_type": "polygon" if anno.seg_mode else "rectangle"
            }

            if anno.seg_mode_polys:
                shape["points"] = anno.polys.exterior.tolist()
            elif anno.seg_mode_mask:
                shape["points"] = anno.mask.to_ploygons().exterior.tolist()
            else:
                shape["points"] = [[anno.bbox.xmin, anno.bbox.ymin], [anno.bbox.xmax, anno.bbox.ymax]]

            anno_json["shapes"].append(shape)

        json.dump(anno_json, open(anno_path, "w"))

    def vis(self, id=None, with_bbox=True, with_seg=True, is_show=False, save_dir=None, reset_dir=False):
        if save_dir is not None:
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            elif reset_dir:
                shutil.rmtree(save_dir)
                os.makedirs(save_dir)

        if id is not None:
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            return sample.vis(with_bbox=with_bbox, with_seg=with_seg, is_show=is_show, save_path=save_path)

        image_vis = []
        for id in tqdm(self.ids):
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            image = sample.vis(with_bbox=with_bbox, with_seg=with_seg, is_show=False, save_path=save_path)
            image_vis.append(image)
        return image_vis



# -*- coding: UTF-8 -*-
import numpy as np
import os
import base64
from ...utils import json_encode
from ...core.http.methods import HttpMethod
from ..process.detect_processor import DetectionProcessor
from ..result import DetectionResult
from .. import detector
from bottle import template, Bottle, request, static_file, tob
from io import BytesIO
import tempfile
import bottle

bottle.BaseRequest.MEMFILE_MAX = 1024 * 1024 * 1024

ERROR_AUTH = 100
ERROR_JSON_PARSER = 101
ERROR_PARAM_LACK = 102
ERROR_IMAGE_PARSER = 103
ERROR_IMAGE_FORMAT = 104
ERROR_DETECT_ERROR = 105

RESOURCE_ROOT = os.path.dirname(detector.__file__) + "/resource"


class DetectorServer(object):
    html_path = RESOURCE_ROOT + "/serve.html"
    static_root = RESOURCE_ROOT + "/static"
    server_name = "Detector Server"
    default_name = "Icv Predictor"

    def __init__(
            self,
            detector,
            secrect=None,
            open_web=True,
            name="",
            default_params_score_thr=0.5):
        self.detector = detector
        self.secrect = secrect
        self.open_web = open_web
        self.app = Bottle()
        self.name = self.default_name if name == "" else ""
        self.default_params_score_thr = default_params_score_thr
        self._fake_req_env()

    def _fake_req_env(self):
        body = "abc"
        request.environ['CONTENT_LENGTH'] = str(len(tob(body)))
        request.environ['wsgi.input'] = BytesIO()
        request.environ['wsgi.input'].write(tob(body))
        request.environ['wsgi.input'].seek(0)

    def _json_response(self, code, message="", success=False, data=None, **kwargs):
        if data is not None:
            res = {
                "code": code,
                "message": message,
                "success": success,
                "data": json_encode(data)
            }
        else:
            res = {
                "code": code,
                "message": message,
                "success": success,
            }

        for k in kwargs:
            res[k] = kwargs[k]

        return res

    def _decrypt(self, encoded):
        return str(base64.b64decode(encoded).decode("utf-8"))

    def _inference_server(self):
        request_data = request.json
        request_headers = request.headers
        try:
            if self.secrect is not None:
                if "secret" in request_headers and self._decrypt(request_headers["secret"]) == self.secrect:
                    pass
                elif "Secret" in request_headers and self._decrypt(request_headers["Secret"]) == self.secrect:
                    pass
                elif "HTTP_SECRET" in request_headers and self._decrypt(request_headers["HTTP_SECRET"]) == self.secrect:
                    pass
                else:
                    return self._json_response(ERROR_AUTH, "Authentication failed.")
        except:
            return self._json_response(ERROR_AUTH, "Authentication failed.")

        if request_data is None:
            return self._json_response(ERROR_JSON_PARSER, 'bad request, json parse failed.')

        if "image" not in request_data:
            return self._json_response(ERROR_PARAM_LACK, "image param can't empty.")

        request_image = request_data["image"]
        try:
            if isinstance(request_image, list):
                image_np = [_ for _ in DetectionProcessor.pre_process(request_image)]
            else:
                image_np = DetectionProcessor.pre_process(request_image)
                if not isinstance(image_np, np.ndarray):
                    return self._json_response(ERROR_IMAGE_FORMAT,
                                               "parse input image failed, please check input format.")
        except Exception as e:
            return self._json_response(ERROR_IMAGE_PARSER, 'bad request, input image parse failed:' + str(e))

        try:
            if isinstance(image_np, list):
                outputs = self.detector.inference_batch(image_np)
            else:
                outputs = self.detector.inference(image_np)
        except Exception as e:
            return self._json_response(ERROR_DETECT_ERROR,
                                       'predict failure, {0}:{1}'.format(e.__class__.__name__, str(e)))

        outputs = DetectionProcessor.post_process(outputs)
        return self._json_response(0, "success", True, outputs)

    def _inference_page(self):
        return template(
            self.html_path,
            name=self.name,
            default_params_score_thr=self.default_params_score_thr
        )

    def _inference_upload_for_predict(self):
        score_thr = request.forms.get("score_thr")

        if score_thr != "":
            score_thr = float(score_thr)
        else:
            score_thr = -1

        upload = request.files.get('uploadfile')
        name, ext = os.path.splitext(upload.filename)
        ext = ext.lower()
        if ext not in ('.png', '.jpg', '.jpeg', ".tif", ".bmp"):
            return self._json_response(ERROR_IMAGE_FORMAT, "Image File Is Invalid!")

        upload.save(self.upload_dir, overwrite=True)
        det_img_name = "{}_det{}".format(name, ext)
        det_img_path = os.path.join(self.upload_dir, det_img_name)
        upload_img_path = os.path.join(self.upload_dir, upload.filename)
        inference_result = self.detector.inference(upload_img_path, save_path=det_img_path, score_thr=score_thr)
        if isinstance(inference_result, DetectionResult):
            return self._json_response(
                code=ERROR_DETECT_ERROR,
                success=True,
                data={
                    "最大置信度": "%.4f" % inference_result.topk(1)[0][1],
                    "检测结果数": inference_result.length,
                    "预测时间": "%.4fs" % inference_result.det_time,
                },
                image_origin="/img/%s" % upload.filename,
                image_detect="/img/%s" % det_img_name,
            )

        return self._json_response(ERROR_DETECT_ERROR, "Predict Error!")

    def start_server(self, port, upload_dir=None, debug=True):
        self.detector._warmup()

        if upload_dir is None:
            upload_dir = tempfile.mkdtemp()

        if not os.path.exists(upload_dir):
            os.makedirs(upload_dir)

        if not os.path.exists(upload_dir):
            raise Exception("Can not create upload directory!")

        self.upload_dir = upload_dir

        self.app.route("/static/:path#.+#", HttpMethod.GET, lambda path: static_file(path, self.static_root))
        self.app.route("/img/:filename", HttpMethod.GET, lambda filename: static_file(filename, upload_dir))
        self.app.route("/upload", HttpMethod.POST, self._inference_upload_for_predict)
        self.app.route("/api", HttpMethod.POST, self._inference_server)
        if self.open_web:
            self.app.route("/page", HttpMethod.GET)(self._inference_page)
            self.app.route("/", HttpMethod.GET)(self._inference_page)

        self.app.run(host="0.0.0.0", port=port, debug=debug)



# -*- coding: UTF-8 -*-
from ...image import imshow, imwrite, imshow_bboxes, immerge
from ...utils import do_post, is_file, is_seq, is_str, labelmap_to_categories, json_decode
import os
import base64

class DetectorClient(object):
    def __init__(self, host, port, secret=None, timeout=3000, categories=None):
        host = host.strip("/")
        self._server_addr = "{}:{}".format(host, port) if str(host).startswith("http://") else "http://{}:{}".format(
            host, port)
        self.rest_url = self._server_addr + "/api"
        self.secret = secret
        self.timeout = timeout

        self.categories = categories
        self.id2cat = None
        self.cat2id = None
        if is_seq(categories) and len(categories) > 0:
            self.id2cat = {ix + 1: cat for ix, cat in enumerate(self.categories)}
            self.cat2id = {cat: ix + 1 for ix, cat in enumerate(self.categories)}

    def _encrypt(self, secret):
        if isinstance(secret, str):
            return base64.b64encode(secret.encode("utf-8"))
        return base64.b64encode(secret)

    def set_categories(self, categories):
        if is_seq(categories) and len(categories) > 0:
            self.categories = list(self.categories)
            self.id2cat = {ix + 1: cat for ix, cat in enumerate(self.categories)}
            self.cat2id = {cat: ix + 1 for ix, cat in enumerate(self.categories)}
        else:
            self.categories = None

    def set_labelmap_path(self, labelmap_path):
        assert is_file(labelmap_path)
        categories = labelmap_to_categories(labelmap_path)
        self.categories = [cat["name"] for cat in categories]
        self.id2cat = {ix + 1: cat for ix, cat in enumerate(self.categories)}
        self.cat2id = {cat: ix + 1 for ix, cat in enumerate(self.categories)}

    def inference(self, image_file, timeout=0, is_show=False, save_path=None, vis_merge=False, vis_merge_origin="x"):
        assert is_file(image_file)
        timeout = timeout if timeout > 0 else self.timeout
        headers = {}
        if self.secret is not None:
            headers["secret"] = self._encrypt(self.secret)

        with open(image_file, "rb") as f:
            base64_data = base64.b64encode(f.read())
            image_encoded = str(base64_data)[2:-1]

        post_data = {
            "image": image_encoded,
        }

        print("ヽ( ^∀^)ﾉ\tinference begin")
        response = do_post(self.rest_url, data=post_data, json=True, headers=headers, timeout=timeout)
        print("ヽ( ^∀^)ﾉ\tinference finish, get response")

        if response.ok:
            try:
                result = json_decode(response.content)
                if result["success"]:
                    result_data = result["data"]
                    if is_str(result_data):
                        result_data = json_decode(result_data.replace("'", "\""))
                    bboxes = result_data["bboxes"]
                    classes = result_data["classes"]
                    scores = result_data["scores"]
                    # TODO: imshow_bboxes过滤分值有错误，待修复
                    # scores = None
                    labels = [self.id2cat[cls] for cls in classes] if self.id2cat is not None else None
                    vis_image = imshow_bboxes(image_file, bboxes, classes=self.categories, labels=labels, scores=scores)

                    if vis_merge:
                        vis_image = immerge([image_file, vis_image], origin=vis_merge_origin)

                    if is_show:
                        imshow(vis_image)
                    if save_path is not None:
                        imwrite(vis_image, save_path)
                    return result["data"], vis_image
                else:
                    print("(´-ι_-｀)\t inference failed! code:(%d) error:(%s)" % (result["code"], result["message"]))
                    return None, None
            except Exception as e:
                print("(థฺˇ౪ˇథ)\tparse result error: ", str(e))
                return None, None
        else:
            print("(థฺˇ౪ˇథ)\tinference error: ", response.error)
            return None, None

    def inference_batch(self, image_file_list, timeout=0, save_dir=None, vis_merge=False, vis_merge_origin="x"):
        for image_file in image_file_list:
            assert is_file(image_file), "image is not exist: {}".format(image_file)

        timeout = timeout if timeout > 0 else self.timeout
        headers = {}
        if self.secret is not None:
            headers["secret"] = self._encrypt(self.secret)

        image_encoded_list = []
        for image_file in image_file_list:
            with open(image_file, "rb") as f:
                base64_data = base64.b64encode(f.read())
                image_encoded = str(base64_data)[2:-1]
                image_encoded_list.append(image_encoded)

        post_data = {
            "image": image_encoded_list,
        }

        response = do_post(self.rest_url, data=post_data, json=True, headers=headers, timeout=timeout)

        if response.ok:
            try:
                result = json_decode(response.content)
                if result["success"]:
                    result_data = result["data"]
                    vis_image_list = []
                    for image_file, data in zip((image_file_list, result_data)):
                        bboxes = data["bboxes"]
                        classes = data["classes"]
                        # scores = result["data"]["scores"]
                        # TODO: imshow_bboxes过滤分值有错误，待修复
                        scores = None
                        labels = [self.id2cat[cls] for cls in classes] if self.id2cat else None
                        vis_image = imshow_bboxes(image_file, bboxes, classes=self.categories, labels=labels,
                                                  scores=scores)

                        if vis_merge:
                            vis_image = immerge([image_file, vis_image], origin=vis_merge_origin)

                        if save_dir is not None:
                            imwrite(vis_image, os.path.join(save_dir, os.path.basename(image_file)))
                    return result_data, vis_image_list
                else:
                    print("(´-ι_-｀)\t inference failed! code:(%d) error:(%s)" % (result["code"], result["message"]))
                    return None, None
            except Exception as e:
                print("(థฺˇ౪ˇథ)\tparse result error: ", str(e))
                return None, None
        else:
            print("(థฺˇ౪ˇథ)\tinference error: ", response.error)
            return None, None



# -*- coding: UTF-8 -*-
from icv.image.io import imread
import os


class Image(object):
    def __init__(self, path=None, img=None, ext=None, name=None, category=None):
        self.path = path
        self.img = img
        self.name = name
        self.category = category
        self.ext = ext

        if self.img is None and self.path is not None:
            self.img = imread(path)

        if self.name is None and self.path is not None:
            self.name = os.path.basename(name)

        if self.ext is None and self.path is not None:
            _, self.ext = os.path.splitext(os.path.split(self.path)[1])

        self.shape = None if self.img is None else self.img.shape
        self.height = self.shape[0] if self.shape is not None else -1
        self.width = self.shape[1] if self.shape is not None else -1
        self.size = (self.height, self.width)



# -*- coding: utf-8 -* -
from .processor import Processor
from ...utils import base64_to_np

class DetectionProcessor(Processor):
    @classmethod
    def pre_process(self, inputs):
        return base64_to_np(inputs)

    @classmethod
    def post_process(self, outputs):
        if isinstance(outputs,list):
            outputs = [_.to_json() for _ in outputs]
        else:
            outputs = outputs.to_json()
        return outputs



# -*- coding: UTF-8 -*-
from .io import imread, imwrite
from skimage import draw
from ..utils import is_seq, is_np_array
from ..vis.color import STANDARD_COLORS, get_color_tuple, get_reverse_color
import random
import numpy as np
import matplotlib.pyplot as plt
import cv2

TEXT_MARGIN = 2

def imshow(img):
    """Show an image.
    Args:
        img (str or ndarray): The image to be displayed.
    """
    plt.axis('off')
    plt.imshow(imread(img))
    plt.show()


def imshow_bboxes(
        img,
        bboxes,
        labels=None,
        scores=None,
        classes=None,
        score_thresh=0,
        masks=None,
        color="red",
        thickness=1,
        use_normalized_coordinates=False,
        is_show=False,
        save_path=None):
    assert classes is None or is_seq(classes) or is_np_array(classes)
    assert labels is None or is_seq(labels) or is_np_array(labels)
    assert scores is None or is_seq(scores) or is_np_array(scores)
    assert masks is None or is_seq(masks) or is_np_array(masks)

    from icv.data.core import BBoxList, BBox

    if isinstance(bboxes, BBoxList):
        bboxes = np.array(bboxes.tolist())
    elif isinstance(bboxes, list) and len(bboxes) > 0 and isinstance(bboxes[0], BBox):
        bboxes = np.array([bbox.bbox for bbox in bboxes])
    else:
        bboxes = np.array(bboxes)

    image = imread(img)

    if bboxes.shape[0] == 0:
        if is_show:
            imshow(image)

        if save_path:
            imwrite(image, save_path)

        return image

    bboxes = bboxes[np.where(np.array(scores) >= score_thresh)] if scores is not None else bboxes

    if labels is not None:
        if not is_np_array(labels):
            labels = np.array(labels)
        labels = labels[np.where(np.array(scores) >= score_thresh)] if scores is not None else labels
        assert labels.shape[0] == bboxes.shape[0], "param labels's length is not equals to bboxes!"

    if masks is not None:
        if not is_np_array(masks):
            masks = np.array(masks)
        masks = masks[np.where(np.array(scores) >= score_thresh)] if scores is not None else masks
        assert masks.shape[0] == bboxes.shape[0], "param masks's length is not equals to bboxes!"

    if scores is not None:
        if not is_np_array(scores):
            scores = np.array(scores)
        scores = scores[np.where(scores >= score_thresh)]

        assert scores.shape[0] == bboxes.shape[0], "param scores's length is not equals to bboxes!"

    colorMap = {}
    default_color = color if color else "DarkOrange"
    if classes is not None:
        if is_np_array(classes):
            classes = list(classes)
        for cat in classes:
            colorMap[cat] = STANDARD_COLORS[classes.index(cat) % len(STANDARD_COLORS)]

    for ix, bbox in enumerate(bboxes):
        if isinstance(bbox, BBox):
            label = labels[ix] if labels is not None else bbox.lable
            color = colorMap[label] if label in colorMap else default_color
            label = label if label is not None else ""
            label = label + ": " + str(round(scores[ix], 3)) if scores is not None else label
        else:
            label = labels[ix] if labels is not None else ""
            color = colorMap[label] if label in colorMap else default_color
            label = label + ": " + str(round(scores[ix], 3)) if scores is not None else label

        if masks is not None:
            image = imdraw_mask(image, masks[ix], color=color)

        if isinstance(bbox, BBox):
            image = imdraw_bbox(image, bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax, color, thickness, label,
                                    use_normalized_coordinates)
        else:
            image = imdraw_bbox(image, bbox[0], bbox[1], bbox[2], bbox[3], color, thickness, label,
                                    use_normalized_coordinates)

    if is_show:
        imshow(image)

    if save_path:
        imwrite(image, save_path)

    return image


def imdraw_bbox(image, xmin, ymin, xmax, ymax, color="red", thickness=1, display_str="", text_color=None,
                use_normalized_coordinates=False):
    assert xmin <= xmax, "xmin shouldn't be langer than xmax!"
    assert ymin <= ymax, "ymin shouldn't be langer than ymax!"

    image = imread(image)
    image = image.copy()

    if xmin == xmax or ymin == ymax:
        return image

    im_height, im_width = image.shape[:2]
    if use_normalized_coordinates:
        (left, right, top, bottom) = (xmin * im_width, xmax * im_width,
                                      ymin * im_height, ymax * im_height)
    else:
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)

    (left, right, top, bottom) = int(left),int(right),int(top),int(bottom)

    color = get_color_tuple(color)

    cv2.rectangle(image, (left, top), (right, bottom), color, thickness=thickness)
    if display_str == "":
        return image

    text_width, text_height,line_height = _get_text_size(display_str)
    text_left = left + TEXT_MARGIN
    text_top = top - TEXT_MARGIN

    if text_top < TEXT_MARGIN:
        text_top = bottom + TEXT_MARGIN

    if text_top + text_height + TEXT_MARGIN > im_height:
        text_top = top + TEXT_MARGIN

    text_color = get_color_tuple(text_color) if text_color is not None else get_reverse_color(color)
    image = imdraw_text(image, display_str, text_left, text_top, text_color=text_color, bg_color=color)
    return image


def imdraw_mask(image, mask, color='red', alpha=0.8):
    """Draws mask on an image.

    Args:
      image: uint8 numpy array with shape (img_height, img_height, 3)
      mask: a uint8 numpy array of shape (img_height, img_height) with
        values between either 0 or 1.
      color: color to draw the keypoints with. Default is red.
      alpha: transparency value between 0 and 1. (default: 0.4)

    Raises:
      ValueError: On incorrect data type for image or masks.
    """
    if mask.dtype != np.uint8:
        raise ValueError('`mask` not of type np.uint8')
    if np.any(np.logical_and(mask != 1, mask != 0)):
        raise ValueError('`mask` elements should be in [0, 1]')

    image = imread(image)

    if image.shape[:2] != mask.shape:
        raise ValueError('The image has spatial dimensions %s but the mask has '
                         'dimensions %s' % (image.shape[:2], mask.shape))

    color_mask = get_color_tuple(color, alpha=alpha)
    color_mask = np.array(color_mask, dtype=np.uint8)

    mask_bin = mask.astype(np.bool)
    image[mask_bin] = image[mask_bin] * alpha + color_mask * (1 - alpha)

    return image


def imdraw_polygons(image, polygons, color='red', alpha=0.8):
    image = imread(image)
    polygons = np.array(_format_polygons(polygons))
    polygons = np.squeeze(polygons)
    X, Y = polygons[:, 0], polygons[:, 1]
    rr, cc = draw.polygon(Y, X)

    rgb = get_color_tuple(color)
    draw.set_color(image, [rr, cc], color=rgb, alpha=alpha)

    return image


def _format_polygons(polygons):
    if is_np_array(polygons):
        polygons = polygons.tolist()
    ps = []
    for polygon in polygons:
        if is_seq(polygon[0]):
            ps.append([tuple(p) for p in polygon])
        else:
            if len(polygon) % 2 == 0:
                ps.append([(polygon[i], polygon[i + 1]) for i in range(len(polygon)) if i % 2 == 0])
    return ps


def _get_bbox_from_points(polygon):
    """
    根据polygon顶点获取bbox
    :param polygon:
    :return:
    """
    x_list = [p[0] for p in polygon]
    y_list = [p[1] for p in polygon]

    xmin = min(x_list)
    ymin = min(y_list)
    xmax = max(x_list)
    ymax = max(y_list)

    return xmin, ymin, xmax, ymax


def imdraw_polygons_with_bbox(
        image,
        polygons,
        label_list=None,
        is_show=False,
        save_path=None,
        alpha=0.5,
        outline=1,
        color_map=None,
        with_bbox=False,
        bbox_color=None,
        text_color=None
):
    image = imread(image)
    polygons = _format_polygons(polygons)

    if len(polygons) <= 0:
        if save_path is not None:
            imwrite(image, save_path)

        if is_show:
            imshow(image)
        return image

    assert label_list is None or len(polygons) == len(label_list)
    assert color_map is None or isinstance(color_map, dict)

    if isinstance(bbox_color, list):
        assert len(bbox_color) == len(polygons)
    if isinstance(text_color, list):
        assert len(text_color) == len(polygons)

    if bbox_color is None:
        bbox_color = random.choice(STANDARD_COLORS)

    if text_color is None:
        text_color = get_reverse_color(bbox_color)

    if label_list is None:
        label_list = [""] * len(polygons)

    tcolor_color_map = {}
    bcolor_color_map = {}
    color_list = []
    for ix, (polygon, label) in enumerate(list(zip(polygons, label_list))):
        tcolor = tcolor_color_map[label] if label in tcolor_color_map else None
        if tcolor is None:
            if isinstance(text_color, list):
                tcolor = text_color[ix]
            else:
                tcolor = text_color

        if label not in tcolor_color_map and tcolor is not None:
            tcolor_color_map[label] = tcolor

        bcolor = bcolor_color_map[label] if label in bcolor_color_map else None
        if bcolor is None:
            if isinstance(bbox_color, list):
                bcolor = bbox_color[ix]
            else:
                bcolor = bbox_color

        if label not in bcolor_color_map and bcolor is not None:
            bcolor_color_map[label] = bcolor

        points = [tuple(p) for p in polygon]
        xmin, ymin, xmax, ymax = _get_bbox_from_points(points)
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)

        image = imdraw_polygons(image, polygon, color=bbox_color, alpha=alpha)

        if with_bbox:
            image = imdraw_bbox(image, left, top, right, bottom, color=bbox_color, thickness=outline, display_str=label,
                                text_color=tcolor)

        color_list.append(bcolor)

    if save_path is not None:
        imwrite(image, save_path)

    if is_show:
        imshow(image)

    return image


def _get_text_size(text, font_face=cv2.FONT_HERSHEY_SIMPLEX, font_scale=0.5,thickness=1):
    size = cv2.getTextSize(text, font_face, font_scale, thickness)
    text_width = size[0][0]
    text_height = size[0][1]
    return text_width, text_height,size[1]


def imdraw_text(image, text="-", x=0, y=0, font_scale=0.5, thickness=1, text_color=(255, 255,255), with_bg=True,
                bg_color="white",bg_alpha=60):
    image = imread(image)
    (height, width) = image.shape[:2]
    text_color = get_color_tuple(text_color)
    if with_bg:
        bg_color = get_color_tuple(bg_color, bg_alpha)
        text_width, text_height,line_height = _get_text_size(text,font_scale=font_scale,thickness=thickness)
        xmin = int(max(0, x - TEXT_MARGIN))
        ymin = int(max(0, y - TEXT_MARGIN - text_height))
        xmax = int(min(width, x + text_width + TEXT_MARGIN))
        ymax = int(min(width, y + TEXT_MARGIN))

        cv2.rectangle(image, (xmin, ymin), (xmax, ymax), bg_color, thickness=-1)
    cv2.putText(image, text, (x, y), cv2.FONT_HERSHEY_SIMPLEX, font_scale, text_color, thickness=thickness)
    return image



# -*- coding: UTF-8 -*-
from ..status_code import HTTP_STATUS_CODES
import cgi
from urllib.parse import urlparse,unquote_plus,parse_qs
from icv.utils import EasyDict
from ..methods import HttpMethod
from ..exceptions import HttpParseException
import json

class HttpRequest(object):
    method = ""
    headers = EasyDict()
    _raw_path = ""
    path = ""
    params = EasyDict()
    query = EasyDict()
    _data = None
    data_post = None
    data_json = None
    requestvars = cgi.FieldStorage()

    def __init__(self,method,headers):
        assert method in HttpMethod.methods
        self.method = method
        self.headers = EasyDict(headers) if isinstance(headers,dict) else EasyDict()

    @property
    def content_type(self):
        if "Content-Type" in self.headers:
            return str(self.headers["Content-Type"]).lower()

        if "content-type" in self.headers:
            return str(self.headers).lower()

        return ""

    @property
    def raw_path(self):
        return self._raw_path

    @raw_path.setter
    def raw_path(self,raw_path):
        self._raw_path = raw_path
        self._parse_url()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self,data):
        self._data = data.decode("utf-8") if isinstance(data,bytes) else data
        if isinstance(self._data, str):
            try:
                self.data_post = json.loads(self._data)
                self.data_json = self.data_post
            except:
                if self._data != "":
                    d = self._build_params(self._data)
                    self.data_post = d if len(d) > 0 else self._data
                else:
                    self.data_post = self._data
                    self.data_json = {}
        else:
            if isinstance(self._data,dict):
                self.data_json = self._data
            self.data_post = self._data

        if "application/json" in self.content_type:
            if self.data_json is None:
                raise HttpParseException("request format error! it's not a valid json format!",400)

    def _build_params(self,params_str):
        if not isinstance(params_str,str):
            return EasyDict()
        try:
            return EasyDict(dict(unquote_plus(y).split('=') for y in params_str.split('&')))
        except:
            return EasyDict()

    def _parse_url(self):
        up = urlparse(self.raw_path)
        self.path = up.path
        self.query = self._build_params(up.query)
        self.params = self._build_params(up.params)

    def __str__(self):
        s = ""
        for attr in self.__dict__:
            s += "{}={}\n".format(attr,self.__dict__[attr])
        return s


class HttpResponse(object):
    def __init__(self,handler):
        self._handler = handler
        self.headers = {}
        self.status = ()

    def add_header(self,header_key,header_value):
        self.headers[header_key] = header_value

    def set_status(self,stat):
        self.status = str(stat)

    def send(self,data,code=200,headers=None):
        headers = headers.update(self.headers) if headers else {}
        return self._handler.response(data, code, **headers)

    def render(self, html, enc="utf-8"):
        return self._handler.render(html, enc)

    def redirect(self,location):
        self.set_status(HTTP_STATUS_CODES[301][0])
        self.add_header("Location", location)

    def __str__(self):
        s = ""
        for attr in self.__dict__:
            s += "{}={}\n".format(attr,self.__dict__[attr])
        return s

class Context(object):
    def __init__(self,handler,req_method,req_headers):
        self.request = HttpRequest(req_method,req_headers)
        self.response = HttpResponse(handler)

    def __repr__(self):
        return "--<Context>--\nrequest=[%s]\nresponse=[%s]" % (str(self.request),str(self.response))

    def __str__(self):
        return self.__repr__()


# -*- coding: UTF-8 -*-
from ..core.http.client.client import IcvHttpClient
from ..core.http.methods import HttpMethod
from .itis import is_dir,is_seq
from .codec import json_encode

def request(method,url,data=None,params=None,headers=None,timeout=3000):
    method = str(method).upper()
    if method == HttpMethod.POST:
        return do_post(url=url,data=data,params=params,headers=headers,timeout=timeout)
    elif method == HttpMethod.GET:
        return do_get(url=url,params=params,headers=headers,timeout=timeout)
    elif method == HttpMethod.PUT:
        return do_put(url=url,data=data,params=params,headers=headers,timeout=timeout)
    elif method == HttpMethod.DELETE:
        return do_post(url=url,data=data,params=params,headers=headers,timeout=timeout)
    elif method == HttpMethod.OPTIONS:
        return do_post(url=url,data=data,params=params,headers=headers,timeout=timeout)
    else:
        raise Exception("Request Method Is Not Support!")

def do_post(url,data=None,json=False,form=False,params=None,headers=None,timeout=3000):
    http_client = IcvHttpClient(url,data=data,params=params,headers=headers,timeout=timeout)
    return http_client.post(json=json,form=form)

def do_get(url,params=None,headers=None,timeout=3000):
    http_client = IcvHttpClient(url, params=params, headers=headers, timeout=timeout)
    return http_client.get()

def do_put(url,data=None,json=False,form=False,params=None,headers=None,timeout=3000):
    http_client = IcvHttpClient(url,data=data,params=params,headers=headers,timeout=timeout)
    return http_client.put(json=json,form=form)

def do_delete(url,data=None,params=None,headers=None,timeout=3000):
    http_client = IcvHttpClient(url, data=data, params=params, headers=headers, timeout=timeout)
    return http_client.delete()

def do_options(url,data=None,params=None,headers=None,timeout=3000):
    http_client = IcvHttpClient(url, data=data, params=params, headers=headers, timeout=timeout)
    return http_client.options()

def file_server(dir,port,url_path_prefix="",debug=False,include_sub=False):
    assert is_dir(dir)
    from bottle import static_file,Bottle
    app = Bottle()
    url_path = "/%s/:path#.+#" % url_path_prefix.strip("/") if include_sub else "/%s/:filename" % url_path_prefix.strip("/")
    app.route(url_path, HttpMethod.GET, lambda path: static_file(path, dir))
    app.run(host="0.0.0.0",port=port,debug=debug)

def simple(port,methods=None,debug=False):
    assert methods is None or is_seq(methods) or len([m for m in methods if m not in HttpMethod.methods])
    if methods is None:
        methods = HttpMethod.methods

    from bottle import Bottle,request
    app = Bottle()

    def _handler(uri):
        return json_encode(dict(
            uri=uri,
            url_path=request.url,
            url_query=request.query.decode(),
            request_method=request.method,
            request_data=request.json,
            request_headers=request.headers
        ))

    app.route(["/<uri:re:.+>", "/"],
              methods,
              _handler)

    app.run(host="0.0.0.0", port=port, debug=debug)
